/**************************************************************************************************
Copyright (c) 2005-2006 Darrell Tam
email: darrell.barrell@gmail.com

Effect main class

General algorithm is:
1. Delay input into input buffer (_x0)
2. If we can interpolate the params or we're forced to output data FFT data from _x0 to _x1
3. run an effect from FFTFx()
4. IFFT data from _x1 to _x2
5. mix data from _x2 into _x3 (output buffer), doing a cross fade with any overlap of data
   already in there

History
	Date       Version    Programmer         Comments
	16/2/03    -          Darrell Tam		 Created buzz version
    01/10/05   1.0        Darrell Tam        Ported to VST

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

***************************************************************************************************/

#include <stdio.h>
#include <string.h>
#include <dt/cplx.h>
#include <dt/WrapProcessFloatVec.h>

#include "rfftw_float.h"
#include "DtBlkFx.hpp"
#include "AEffEditor.hpp"
#include "Gui.h"
#include "FFTFx.h"

//#define DBG_FILE_NAME "c:\\blkfx1.txt"
#include "Debug.h"

using namespace std;

// constants
enum {
	AUDIO_CHANNELS = BlkFxParam::AUDIO_CHANNELS
};
//-------------------------------------------------------------------------------------------------

DtBlkFx::DtBlkFx (audioMasterCallback audioMaster)
	: AudioEffectX (audioMaster, /*kNumPrograms*/g_num_blk_fx_presets, /*kNumParams*/BlkFxParam::TOTAL_NUM)
{	
	DBG("DtBlkFx constructor");

	int i;

	// get buffer space
	_x0_sz = 8*44100;								// length of buffer is arbitrary
	_x0_force_out_sz = _x0_sz - MAX_FFT_BLK_SZ;		// force output when x0 contains this much data
	for(i = 0; i < AUDIO_CHANNELS; i++) {
		_chan[i].x0.resize(_x0_sz+MAX_FFT_BLK_SZ);	// input buffer with extra space at end to unwrap data for processing
		_chan[i].x1.resize(MAX_FFT_BLK_SZ+1);		// allow 1 extra for imaginary part of DC (which is always 0)
		_chan[i].x3.resize(5*44100);				// output buffer (length is arbitrary, as long as > MAX_FFT_BLK_SZ plus a few)
	}
	_max_delay_n = _chan[0].x3.size()-MAX_FFT_BLK_SZ-2048;

	_x2.resize(MAX_FFT_BLK_SZ+1);	// inverse FFT

	for(i = 0; i < BlkFxParam::NUM_FX_SETS; i++) {
		// allocate 10k per effect 
		_fx_params[i].data.resize(10000);

		_fx_params[i].fft_fx = NULL;
	}

	// put in arbitrary initial params - they should be overwritten by first param update
	_params.put(0, BlkFxParam::DELAY, 16.0f/255.0f);
	_params.put(0, BlkFxParam::FFT_LEN, BlkFxParam::getFFTLenParam(5));
	_params.put(0, BlkFxParam::OVERLAP, 0.35f);

	// copy presets into the program
	_program.resize(AudioEffect::numPrograms);
	for(i = 0; i < AudioEffect::numPrograms; i++) _program[i] = g_blk_fx_presets[i];

	setNumInputs (AUDIO_CHANNELS);
	setNumOutputs(AUDIO_CHANNELS);

	hasVu ();
	hasClip(false);
	canProcessReplacing(true);
	programsAreChunks(true);

	// didn't work in fl studio or renoise for delay of 4096
	_initial_delay = 0;
	setInitialDelay(_initial_delay);

	// these were registered from steinberg
	if(AUDIO_CHANNELS == 2) setUniqueID ('h526');
	else setUniqueID ('g319');

	editor = new Gui(this);

	init();

	resume ();						// flush buffer
}

//-------------------------------------------------------------------------------------------------
DtBlkFx::~DtBlkFx ()
{
}

//-------------------------------------------------------------------------------------------------
void DtBlkFx::init()
// clear out all buffers
{
	unsigned i;

	// current absolute sample position of input/output updated with every input buf
	// i.e. absolute position of _x0[_x0_i+x0_n] / _x3[_x3_o]
	_curr_samp_abs = 0;
	_next_blk_fwd_n = 0;

	// absolute sample position of start of current fft blk at x0[x0_i]/x1[0]/x2[0]
	_blk_samp_abs = 0;
	_x0_n_past_end = 0;		// number of samples copied past end of buffer
	_x0_i = 0;				// pre-fft buffer offset
	_x0_n = 0;				// number of samples in the pre-fft buffer

	_x3_o = 0;				// output FIFO output index
	_x3_end_abs = 0;		// no data in _x3

	// clear out the output buffers
	for(i = 0; i < AUDIO_CHANNELS; i++)
		memset(&(_chan[i].x3[0]), 0, _chan[i].x3.size()*sizeof(float));

	_x3_is_clear = true;

	// these will be updated on first poll
	sampleRate = 44100.0f;
	_prev_poll_abs = 0;
	_samps_per_poll = 0;
	_samps_per_tick = 250.0f;
	_samps_per_beat = 4000.0f;
	_beat_start_abs = 0;
	_min_blk_fwd_samps = 100;

	// cause effects to initialize
	for(i = 0; i < BlkFxParam::NUM_FX_SETS; i++) {
		_fx_params[i].prev_type = -1;
		_fx_params[i].type = -1;
	}

	_params_state = PARAMS_CHK_SYNC;
	_params_need_processing = true;

	// these variables will be updated on first paramsChk()
	_dst_fft_abs = 0;
	_freq_fft_n = 4096;

}

//-------------------------------------------------------------------------------------------------
long DtBlkFx::getChunk (void** data, bool isPreset)
// virtual, override AudioEffect
// only save using the latest chunk version
{
	DBG("DtBlkFx::getChunk" << VAR(isPreset));
	unsigned i;

	// how many programs are we saving? 
	unsigned num_programs = isPreset? 0 : _program.size();

	// get enough space to copy current data and all presets 
	_chunk_data.resize(ChunkDataV1_0::size(num_programs));

	// cast to what we need
	Ptr<ChunkDataV1_0> chunk_data((void*)&_chunk_data[0]);

	chunk_data->tag = chunk_data->TAG;
	chunk_data->version = chunk_data->VERS;
	chunk_data->num_programs = num_programs;

	// save current settings
	chunk_data->curr.clrName();
	for(i = 0; i < BlkFxParam::TOTAL_NUM; i++)
		chunk_data->curr.params[i] = _params.getInput(i);

	// save all the other stuff
	if(isPreset) {
		DBG("DtBlkFx::getChunk saving preset" << VAR(currProgramNum()) << VAR(currProgram().name));

		// save the current preset only
		chunk_data->curr.setName(currProgram().name);
	}
	else {
		// write the current preset number into the name (I did it like this because I forgot to add a field for prerelease)
		char curr_name[32];
		sprintf(curr_name, "%d", currProgramNum());
		chunk_data->curr.setName(curr_name);

		// save all presets
		DBG("DtBlkFx::getChunk saving" << VAR(chunk_data->num_programs));
		for(i = 0; i < chunk_data->num_programs; i++)
			chunk_data->program(i) = _program[i];
	}

	*data = chunk_data;
	return _chunk_data.size(); // bytes
}

//-------------------------------------------------------------------------------------------------
template <class CHUNK_DATA> inline bool ChunkOk(Ptr<CHUNK_DATA> data, long bytes)
// return true if "data" seems to be ok (size, tag & version match)
{
	bool ok =
		bytes >= sizeof(CHUNK_DATA) &&	// make sure there's enough space for the initial part
		data->tag == CHUNK_DATA::TAG &&
		data->version == CHUNK_DATA::VERS &&
		bytes == data->size();			// now we can call the "size" method

	DBG("is ChunkOk?" << VAR(sizeof(CHUNK_DATA)) << VAR(bytes) << VAR(data->tag) <<VAR(CHUNK_DATA::TAG) << VAR(data->version) << VAR(CHUNK_DATA::VERS) << VAR(ok));
	return ok;
}

//-------------------------------------------------------------------------------------------------
long /*1=success*/ DtBlkFx::setChunk (void* data, long byteSize, bool isPreset)
// virtual, override AudioEffect
{

	{
		Ptr<ChunkDataV0_9> chunk_data(data);
		if(ChunkOk(chunk_data, byteSize)) {
			DBG("DtBlkFx::setChunk v0.9" << VAR(byteSize) << VAR(isPreset));
			if(isPreset) return 0;	// preset recall not supported with old version

			for(int i = 0; i < BlkFxParam::TOTAL_NUM; i++) setParameter(i, chunk_data->params[i]);
			return 1;
		}
	}

	{
		Ptr<ChunkDataV1_0> chunk_data(data);
		if(ChunkOk(chunk_data, byteSize)) 
		{
			DBG("DtBlkFx::setChunk v1.0" << VAR(byteSize) << VAR(isPreset) << VAR(chunk_data->num_programs));
			if(isPreset) {
				// copy the "current" params into the current preset
				currProgram() = chunk_data->curr;
				setProgram(currProgramNum());
			}
			else {
				// recall current params
				unsigned i;
				// recall presets
				for(i = 0; i < _program.size() && i < chunk_data->num_programs; i++) {
					_program[i] = chunk_data->program(i);
					DBG("DtBlkFx::setChunk v1.0" << VAR(i) << VAR(_program[i].getName()));
				}

				// retrieve current program number from name of current param
				// (I did it like this because I forgot to add a field for prerelease)
				sscanf(chunk_data->curr.getName(), "%d", &curProgram);
				DBG("DtBlkFx::setChunk v1.0" << VAR(curProgram));
				curProgram = currProgramNum(); // limit range

				for(i = 0; i < BlkFxParam::TOTAL_NUM; i++)
					setParameter(i, chunk_data->curr.params[i]);


			}
			return 1;
		}
	}
	DBG("DtBlkFx::setChunk, unknown type" << VAR(byteSize) << VAR(isPreset));
	return 0;	// can't recall
}


//-------------------------------------------------------------------------------------------------
void DtBlkFx::setProgram (long num)
// virtual, override AudioEffect
// called by vst to set a program
{
	DBG("DtBlkFx::setProgram" << VAR(num));

	// check for the special "reset current program"
	if(num < (int)_program.size()-1) AudioEffect::setProgram(num); // normal case
	else {
		// reset current program to defaults
		currProgram() = g_blk_fx_presets[currProgramNum()];
	}

	// do the select
	for(int i = 0; i < BlkFxParam::TOTAL_NUM; i++) setParameter(i, currProgram().params[i]);
}

//-------------------------------------------------------------------------------------------------
void DtBlkFx::setProgramName (char *name)
// virtual, override AudioEffect
// called by vst to change the name of a program
{
	DBG("DtBlkFx::setProgramName" << VAR(name));
	currProgram().setName(name);
}

//-------------------------------------------------------------------------------------------------
void DtBlkFx::getProgramName (char *name)
// virtual, override AudioEffect
// called by vst to get the name of a program
{
	DBG("DtBlkFx::getProgramName" << VAR(currProgram().getName()));
	strcpy(name, currProgram().getName());
}

//-------------------------------------------------------------------------------------------------
bool DtBlkFx::getProgramNameIndexed (long category, long index, char* text)
// virtual, override AudioEffect
// called by vst to get the name of a program by category
{
	DBG("DtBlkFx::getProgramNameIndexed" << VAR(category) << VAR(index) << VAR(_program.size()));
	if(index < 0 || index >= (int)_program.size()) return false;
	strcpy(text, _program[index].name);
	return true;
}

//-------------------------------------------------------------------------------------------------
void DtBlkFx::resume ()
// virtual, override AudioEffect
// called by vst to get ready to start processing
{
	DBG("DtBlkFx::resume");
	AudioEffectX::resume ();
	if(editor) gui()->resume();
}

//-------------------------------------------------------------------------------------------------
void DtBlkFx::suspend ()
// virtual, override AudioEffect
// called by vst when sound is stopping
{
	DBG("DtBlkFx::suspend");
	AudioEffectX::suspend();

	// roll back params to just contain most recent and reset sample position
	_params.resetAndCopyIn();

	// clear buffers
	init();

	if(editor) gui()->suspend();
}

//-------------------------------------------------------------------------------------------------
void DtBlkFx::setBlockSize(long sz)
// virtual, override AudioEffect
// called by vst to indicate max number of samples that will be passed to process
{
	AudioEffectX::setBlockSize(sz);
	_max_delay_n = _chan[0].x3.size()-MAX_FFT_BLK_SZ-sz;
}

struct MyInfo : public VstTimeInfo
{
	void dbgprint() {
		DBG("VstTimeInfo" 
			<< VAR(samplePos)
			<< VAR(sampleRate)
			<< VAR(nanoSeconds)
			<< VAR(ppqPos)
			<< VAR(tempo)
			<< VAR(barStartPos)
			<< VAR(cycleStartPos)
			<< VAR(cycleEndPos)
			<< VAR(timeSigNumerator)
			<< VAR(timeSigDenominator)
			<< VAR(smpteOffset)
			<< VAR(smpteFrameRate)
			<< VAR(samplesToNextClock)
			<< VAR(flags)			
		);
	}
};
//-------------------------------------------------------------------------------------------------
float/*hz*/DtBlkFx::guessRoundHz(float freq_hz, float bin_adjust_frac) const
// given a freq, round using the guessed FFT len
{
	bool neg = false;
	if(freq_hz < 0.0f) {
		freq_hz = -freq_hz;
		neg = true;
	}
	float fft_n = guessFFTLen();
	float hz_per_bin = sampleRate/fft_n;
	float bin = limit_range(freq_hz/hz_per_bin, 0.0f, fft_n/2.0f);
	float r = (floorf(bin+0.5f)+bin_adjust_frac)*hz_per_bin;
	return neg?-r:r;
}

//-------------------------------------------------------------------------------------------------
inline void DtBlkFx::pollUpdate(bool force)
// internal method, called regularly from process methods
//
// update polled variables from vst host (to do with tempo)
//
{
	if(!force && _curr_samp_abs - _prev_poll_abs < _samps_per_poll) return;

	// ask host for tempo
	VstTimeInfo* ti = getTimeInfo(kVstTempoValid|kVstPpqPosValid);
	//((MyInfo*)ti)->dbgprint();

	if(ti->flags & kVstTempoValid) 
		_samps_per_beat = (float)(60.0*sampleRate/ti->tempo);
	
	// calculate samples per tick (based on tempo)
	_samps_per_tick = (1.0f/(float)BlkFxParam::TICKS_PER_BEAT)*_samps_per_beat;

	// interpolate params 1/4 beat before the value changes
	_params.setExpectedDist((long)(_samps_per_beat*.25f));

	float ppq_pos = (float)ti->ppqPos;
	float ppq_frac = ppq_pos-floorf(ppq_pos);

	// work out sample position of the start of the current beat
	if(ti->flags & kVstPpqPosValid)
		_beat_start_abs = _curr_samp_abs-(long)(ppq_frac*_samps_per_beat);

	//DBG("pollUpdate" <<VAR(_samps_per_beat)<<VAR(ppq_pos)<<VAR(ppq_frac)<<VAR(_beat_start_abs));

	// poll every beat
	_samps_per_poll = (long)(_samps_per_beat);
	_prev_poll_abs = _curr_samp_abs;

	// min number of samps we can mv fwd to get the next blk
	_min_blk_fwd_samps = max(100L, (long)_samps_per_tick);
}

//-------------------------------------------------------------------------------------------------
void DtBlkFx::setParameter (long index, float value)
// virtual, override AudioEffect
// called by vst to set param
{
	using namespace BlkFxParam;
	if(index < 0 || index >= TOTAL_NUM) return;

	// copy change into the current program
	currProgram().params[index] = value;

	if(_params.put(_curr_samp_abs, index, value) /*already exists*/ == false) {
		// this is a new time
		pollUpdate(/*force*/true);

		// force processing loop to try re-interpolating the params if it hasn't already
		if(_params_state != PARAMS_INTERP_OK)
			_params_need_processing = true;
	}

	if (editor)
		((AEffGUIEditor*)editor)->setParameter (index, value);
	//if (editor) editor->postUpdate ();
}

//-------------------------------------------------------------------------------------------------
float DtBlkFx::getParameter (long index)
// virtual, override AudioEffect
// called by vst to get param value
{
	using namespace BlkFxParam;
	if(index < 0 || index >= TOTAL_NUM) return 0.0f;
	// use most recently input value
	return _params.getInput(index);
}


//-------------------------------------------------------------------------------------------------
void DtBlkFx::getParameterName (long index, char *label)
// virtual, override AudioEffect
// called by vst to get name of param
{
	using namespace BlkFxParam;
	SplitParamNum p(index);

	// try global params
	switch(p.glob_param) {
	case MIX_BACK: strcpy(label, "Mix back"); return;
	case DELAY: strcpy(label, "Delay"); return;
	case FFT_LEN: strcpy(label, "Blk Len"); return;
	case OVERLAP: strcpy(label, "Overlap"); return;
	}
	// try fx set
	switch(p.fx_param) {
	case FX_FREQ_A: sprintf(label, "%d: Freq A", p.fx_set); return;
	case FX_FREQ_B: sprintf(label, "%d: Freq B", p.fx_set); return;
	case FX_AMP: sprintf(label, "%d: Rel Amp", p.fx_set); return;
	case FX_TYPE: sprintf(label, "%d: Effect", p.fx_set); return;
	case FX_VAL: sprintf(label, "%d: Value", p.fx_set); return;
	}
}

//-------------------------------------------------------------------------------------------------
void DtBlkFx::getParameterLabel (long index, char *label)
// virtual, override AudioEffect
// called by vst to get param label
{
	using namespace BlkFxParam;
	SplitParamNum p(index);
	switch(p.glob_param) {
	case MIX_BACK: strcpy(label, "%"); return;
	case DELAY: strcpy(label, "beats"); return;
	case FFT_LEN: strcpy(label, "sec"); return;
	case OVERLAP: strcpy(label, "%"); return;
	}
	switch(p.fx_param) {
	case FX_FREQ_A: strcpy(label, "Hz"); return;
	case FX_FREQ_B: strcpy(label, "Hz");  return;
	case FX_AMP: strcpy(label, "dB");  return;
	case FX_TYPE: strcpy(label, "");  return;
	case FX_VAL: strcpy(label, "");  return;
	}
}

//-------------------------------------------------------------------------------------------------
void DtBlkFx::getParameterDisplay (long index, char *text)
// virtual, override AudioEffect
// called by vst to convert a param to text
{
	using namespace BlkFxParam;
	SplitParamNum p(index);

	if(!p.ok) {
		sprintf(text, "?");
		return;
	}

	// get current input
	float v = _params.getInput(index);

	// see if index belongs to the global params
	switch(p.glob_param) {
	case MIX_BACK:
		sprintf(text, "%s %2.0f", getPwrMatch(v)?"match":"filter", getMixBackFrac(v)*100.0f); return;

	case DELAY:
		sprintf(text, "%1.3f", (getDelaySamps(v)-_initial_delay)/_samps_per_beat); return;

	case FFT_LEN: {
			float fft_len = guessFFTLen();

			// determine blk size in seconds
			float sec = fft_len/sampleRate;

			// print blk size in seconds or msec
			sprintNum(text, sec);
			return;
		}

	case OVERLAP: {
			float fft_len = guessFFTLen();
			float samps_fwd = (float)getBlkShiftFwd(v, fft_len);
			float overlap = 1.0f-(samps_fwd/fft_len);
			sprintf(text, "%s%.0f", getBlkSync(v)?"sync ":"", overlap*100.0f); return;
		}
	}

	// get the fx interface
	int fx_idx = getEffectType(
		_params.getInput( paramOffs(p.fx_set)+FX_TYPE )
	);
	FFTFx* fft_fx = GetFFTFx(fx_idx);

	switch(p.fx_param) {
	case FX_FREQ_A:
	case FX_FREQ_B: {
			float freq_hz = guessRoundHz(BlkFxParam::getHz(v));
			if(!fft_fx->paramUsed(p.fx_param)) 
				sprintf(text, "not used");
			else {
				// put in note based on freq
				text += HzToNote(text, freq_hz);
				*text++ = '/';
				// print in Hz or kHz
				sprintNum(text, freq_hz, /*min unit*/1.0f);
			}
			return;
		}
	case FX_AMP: {
			bool mix_mode = fft_fx->ampMixMode();
			float amp_mult = getEffectAmpMult(v, mix_mode);
			if(!fft_fx->paramUsed(FX_AMP))
				sprintf(text, "not used");
			else if(mix_mode && amp_mult <= 1.0f)
				sprintf(text, "%.0f %%", amp_mult*100.0f);
			else if(amp_mult > 0)
				sprintNum(text, getEffectAmp(v), /*min unit*/1.0f, /*max unit*/100.0f);
			else
				sprintf(text, "-inf");
			return;
		}

	case FX_TYPE:
		strcpy(text, GetFFTFx(getEffectType(v))->name());
		return;

	case FX_VAL:
		GetFFTFx(fx_idx)->dispVal(
			this,		//
			text,		// output
			v,			// value
			p.fx_set	// which fx set we are
		);
		return;
	}
}


//-------------------------------------------------------------------------------------------------
bool DtBlkFx::getEffectName (char* name)
// virtual, override AudioEffect
// called by vst to return name of effect
{
	DBG("DtBlkFx::getEffectName");
	strcpy (name, "DTBlkFx");
	return true;
}

//-------------------------------------------------------------------------------------------------
bool DtBlkFx::getProductString (char* text)
// virtual, override AudioEffect
// called by vst to get product string
{
	DBG("DtBlkFx::getProductString");
	strcpy (text, "DtBlkFx");
	return true;
}

//-------------------------------------------------------------------------------------------------
bool DtBlkFx::getVendorString (char* text)
// virtual, override AudioEffect
// called by vst to get vendor string
{
	DBG("DtBlkFx::getVendorString");
	strcpy (text, "DT");
	return true;
}


//-------------------------------------------------------------------------------------------------
inline void DtBlkFx::doFFT()
// internal method
//
// perform FFT on all channels, find accumulative power of spectrum & scale to prevent overflow when
// doing effects
//
{
	// scale spectrum
	float scale = 1.0f/_freq_fft_n;

	// do the FFT
	for(int i = 0; i < AUDIO_CHANNELS; i++) {
		Chan& chan = _chan[i];

		rfftw_one(g_plan_fft[_plan], &(chan.x0[_x0_i]), &(chan.x1[0]));

		// set imaginary part of dc to be 0 (so 0hz is now a complex number)
		chan.x1[_freq_fft_n] = 0;

		// find power of spectrum
		float acc = 0.0f; 
		pcplxf_range in(FFTdata(i), /*start*/0, /*end*/_freq_fft_n/2);
		for(; in.r != in.e; in++) {
			*in = *in * scale;
			acc += norm(*in);
		}

		// nyquist freq is real only
		*(in.r) *= scale;
		acc += *(in.r) * *(in.r); 

		chan.total_out_pwr = acc;
		chan.total_in_pwr = acc;
	}
}


//-------------------------------------------------------------------------------------------------
inline void DtBlkFx::procFFT()
// internal method
// process the FFT'd data (x1) using the fx_params
//
{
	float fft_len = (float)_freq_fft_n;

	// collect all parameters first
	for(_fx_set = 0; _fx_set < BlkFxParam::NUM_FX_SETS; _fx_set++) {
		FxParams& p = currFxParams();

		int param_offs = BlkFxParam::paramOffs(_fx_set);
		p.freq_param[0] = _params.getInterp(BlkFxParam::FX_FREQ_A+param_offs);
		p.freq_param[1] = _params.getInterp(BlkFxParam::FX_FREQ_B+param_offs);
		p.amp_param = _params.getInterp(BlkFxParam::FX_AMP+param_offs);
		p.val = _params.getInterp(BlkFxParam::FX_VAL+param_offs);
		long type = BlkFxParam::getEffectType(
			_params.getNonInterp(BlkFxParam::FX_TYPE+param_offs)
		);

		p.prev_type = p.type;
		p.type = type;
		p.fft_fx = GetFFTFx(type);

		// convert the params
		p.fbin[0] = BlkFxParam::getFFTBin(p.freq_param[0], fft_len, sampleRate);
		p.fbin[1] = BlkFxParam::getFFTBin(p.freq_param[1], fft_len, sampleRate);
		p.amp = BlkFxParam::getEffectAmpMult(p.amp_param, p.fft_fx->ampMixMode());

		// round bins to ints
		p.bin[0] = (long)(p.fbin[0]+0.5f);
		p.bin[1] = (long)(p.fbin[1]+0.5f);
	}

	// run each effect

	for(_fx_set = 0; _fx_set < BlkFxParam::NUM_FX_SETS; _fx_set++)
		currFxParams().fft_fx->process(this);

	// post process, work pwr out scaling
	for(unsigned i = 0; i < AUDIO_CHANNELS; i++) {
		Chan* chan = _chan+i;

		// remove extremes because they're annoying
		chan->x1[0] = 0;
		chan->x1[_freq_fft_n/2] = 0;

		// match the output to the input power
		if(_pwr_match) {
			// power match mode, scale output to match input power
			double pwr_scale = 1.0f;
			if(chan->total_out_pwr > 1e-30)
				pwr_scale *= chan->total_in_pwr/chan->total_out_pwr;

			if(pwr_scale > 1e30) pwr_scale = 1.0f;			// too big
			if(pwr_scale < 1e-30) pwr_scale = 0.0f;			// too small (or worse, negative)
			chan->out_pwr_scale = (float)pwr_scale;
			chan->out_scale = (float)sqrt(pwr_scale);
		}
		else {
			// filter mode
			chan->out_pwr_scale = 1.0f;
			chan->out_scale = 1.0f;
		}
	}
}


//-------------------------------------------------------------------------------------------------
inline void DtBlkFx::prepOut()
// internal method
// 
// Prepare to process a blk by finding the number of samples that the current blk overlaps
// with the previous blk (_fadein_n & _fadeout_n)
//
// note that _x3_end_abs is updated, which is slightly early because the data hasn't yet
// been copied in
//
{
	// the amount of xfade-in corresponds to the overlap of this fft blk with existing data
	_fadein_n = _x3_end_abs-_dst_fft_abs;

	// normal case is not to xfade-out current fft-blk because next blk will do the xfade
	_fadeout_n = 0;

	// get the absolute ending position of blk
	long dst_fft_end_abs = _dst_fft_abs+_time_fft_n;

	// how much the new fft blk will increase the amount of data in x3
	long extend_x3 = dst_fft_end_abs - _x3_end_abs;

	if(_fadein_n >= 0) {
		// normal case: previous data and the current fft blk overlap, cross-fade

		if(extend_x3 < 0) {
			// special case: current fft blk ends up entirely within existing x3 data, 

			// do nothing if the blk ends prior to the current output position (should not happen)
			if(_curr_samp_abs - dst_fft_end_abs >= 0) {
				DBG("mixToOutputBuffer possible ERROR" << VAR(_curr_samp_abs - dst_fft_end_abs));
				return;
			}

			// cross fade all of fft blk in & out
			_fadein_n = _time_fft_n/2;
			_fadeout_n = _time_fft_n-_fadein_n;

			// one case is not handled, part of current blk is before current sample position
			// (should never happen)
		}
	}
	else {
		// special case: there is a gap between the end of prev data and the current fft blk

		// zero-fill gap (and a bit extra because we will still cross-fade to the current blk)
		long zero_o = _x3_end_abs-_curr_samp_abs;
		long zero_n = -_fadein_n;

		// arbitrary fade in of current fft blk
		_fadein_n = 16;

		// zero data in the gap (unless x3 is already cleared)
		if(!_x3_is_clear) {
			for(int i = 0; i < AUDIO_CHANNELS; i++)
				wrapProcess(PZero(), _chan[i].x3, _x3_o+zero_o, zero_n+_fadein_n);
		}
		_x3_is_clear = false;

		// arbitrary fade out of previous data 
		long prev_fade_n = 16;
		long prev_fade_o = zero_o-prev_fade_n;

		// don't fade data that has already been output
		if(prev_fade_o < 0) {
			prev_fade_n += prev_fade_o;
			prev_fade_o = 0;
		}
		if(prev_fade_n > 0) {
			for(int i = 0; i < AUDIO_CHANNELS; i++)
				wrapProcess(PFadeOut(prev_fade_n), _chan[i].x3, _x3_o+prev_fade_o, prev_fade_n);
		}
		//DBG("mixToOutputBuffer gap" << VAR(zero_o) << VAR(zero_n) << VAR(prev_fade_n) << VAR(prev_fade_o));
	}
	// update the end position of data in x3 if it is extended
	if(extend_x3 > 0) _x3_end_abs = dst_fft_end_abs;

	// ensure fade in is not too big
	if(_fadein_n > _time_fft_n) _fadein_n = _time_fft_n;
}

//-------------------------------------------------------------------------------------------------
inline void DtBlkFx::procBlk()
//
// do the FFT, all FFT processing, IFFT and mixing to output buffer
//
// mix from x2 (post-IFFT) & x0 (pre-FFT) into x3
// x2*(1-mix_back) + x0*mix_back => x3
//
{
	// offset into x3 at which to cross fade in first sample of fft blk
	long x3_fadein_o = _dst_fft_abs-_curr_samp_abs+_x3_o;

	// offset into x3 at which to start copy of curr fft blk
	long x3_cp_o = x3_fadein_o + _fadein_n;

	// number of samples to copy from fft blk
	long cp_n = _time_fft_n - _fadein_n - _fadeout_n;

	// offset into x3 at which to start fade-out
	long x3_fadeout_o = x3_cp_o+cp_n;

	float mix_back_param = _params.getInterp(BlkFxParam::MIX_BACK);
	float mix_back = limit_range(BlkFxParam::getMixBackFrac(mix_back_param), 0.0f, 1.0f);
	_pwr_match = BlkFxParam::getPwrMatch(mix_back_param);

	// don't do any FFTs if output data is completely original, don't do any FFTs
	if(mix_back >= 1.0f) {
		for(int i = 0; i < AUDIO_CHANNELS; i++) {
			if(_fadein_n)
				wrapProcess(PXFade(&_chan[i].x0[_x0_i], 1.0f, _fadein_n), _chan[i].x3, x3_fadein_o, _fadein_n);
			if(cp_n)
				wrapProcess(PScaleCopy(&_chan[i].x0[_x0_i+_fadein_n], 1.0f), _chan[i].x3, x3_cp_o, cp_n);
			if(_fadeout_n)
				wrapProcess(PXFade(&_chan[i].x0[_x0_i+_fadein_n+cp_n], 1.0f, _fadeout_n, /*rev*/true),
					_chan[i].x3, x3_fadeout_o, _fadeout_n);
		}
		return;
	}

	// normal case, we need to do the FFTs
	doFFT();
	if(editor) gui()->FFTDataRdy(0/*input*/);
	procFFT();
	if(editor) gui()->FFTDataRdy(1/*output*/);

	// do iFFT, then fade-in, direct copy and fade-out to output buffer
	for(int i = 0; i < AUDIO_CHANNELS; i++) {
		Chan& chan = _chan[i];

		// inverse fft
		rfftw_one(g_plan_ifft[_plan], &(chan.x1[0]), &_x2[0]);

		// find processed data scaling
		float x2_scale = (1.0f-mix_back) * chan.out_scale; 

		if(mix_back <= 0.0f) {
			// output data is completely fft blk
			if(_fadein_n)
				wrapProcess(PXFade(&_x2[0], x2_scale, _fadein_n), chan.x3, x3_fadein_o, _fadein_n);
			if(cp_n)
				wrapProcess(PScaleCopy(&_x2[_fadein_n], x2_scale), chan.x3, x3_cp_o, cp_n);
			if(_fadeout_n)
				wrapProcess(PXFade(&_x2[_fadein_n+cp_n], x2_scale, _fadeout_n, /*rev*/true),
					chan.x3, x3_fadeout_o, _fadeout_n);
		}
		else {
			// output data is a mix of original and processed
			if(_fadein_n)
				wrapProcess(PXFade2(&_x2[0], x2_scale, &(chan.x0[_x0_i]), mix_back, _fadein_n),
					chan.x3, x3_fadein_o, _fadein_n);
			if(cp_n)
				wrapProcess(PScaleCopy2(&_x2[_fadein_n], x2_scale, &(chan.x0[_x0_i+_fadein_n]), mix_back),
					chan.x3, x3_cp_o, cp_n);
			if(_fadeout_n)
				wrapProcess(PXFade2(&_x2[_fadein_n+cp_n], x2_scale, &(chan.x0[_x0_i+_fadein_n+cp_n]), mix_back, _fadeout_n, /*rev*/true),
					chan.x3, x3_fadeout_o, _fadeout_n);
		}
	}
}

//-------------------------------------------------------------------------------------------------
inline void DtBlkFx::nextBlk()
// internal method
// update everything ready for the next block
//
{
	float overlap_param = _params.getInterp(BlkFxParam::OVERLAP);

	_next_blk_fwd_n = BlkFxParam::getBlkShiftFwd(overlap_param, (float)_time_fft_n);

	// synchronize next blk to start-of-beat position if blk_sync is true
	if(BlkFxParam::getBlkSync(overlap_param)) {
		// next blk sample position as determined from current overlap param
		long next_blk_samp_abs = _blk_samp_abs+_next_blk_fwd_n;
				
		// find the sample position of the most recent beat before next blk position
		float beats = _samps_per_beat>0.0f ?
			ceilf((_beat_start_abs-next_blk_samp_abs)/_samps_per_beat)
			: 1/*this case should never happen*/;

		long sync_abs = (long)(_beat_start_abs-beats*_samps_per_beat);
		long sync_fwd = sync_abs-_blk_samp_abs;

		//DBG("nextBlk, blksync" <<VAR(_next_blk_fwd_n) <<VAR(sync_fwd) <<VAR(_blk_samp_abs) <<VAR(next_blk_samp_abs)<<VAR(beats)<<VAR(sync_abs)<<VAR(overlap_param));
		if(sync_fwd > _min_blk_fwd_samps && sync_fwd <= _next_blk_fwd_n) _next_blk_fwd_n = sync_fwd;
	}

	_params_state = PARAMS_CHK_SYNC;
	_params_need_processing = true;
}


//-------------------------------------------------------------------------------------------------
inline void DtBlkFx::paramsChk()
// internal method
//
// check whether params need processing & update variables used in processing loop
//
{
	if(!_params_need_processing) return;
	_params_need_processing = false;

	// chk if the next param is available and possibly sync to it
	if(_params_state == PARAMS_CHK_SYNC && _params.setOutPos(_blk_samp_abs+1, /*find fraction*/false)) {

		// should we sync fft blk with param?
		if(BlkFxParam::getBlkSync(_params.getNonInterp(BlkFxParam::OVERLAP, /*next param*/true)) && 
			_params.isAnyExplicitSet(/*next param*/true)) {

			// find abs sample position of param
			long param_abs = _params.getSampAbs(/*next param*/true);

			// how far ahead of the current position is this?
			long param_fwd = param_abs-_blk_samp_abs;

			//DBG("paramsChk, blksync" <<VAR(_params.getSampAbs(false)) <<VAR(param_abs) <<VAR(param_fwd) <<VAR(_next_blk_fwd_n));
			if(param_fwd > _min_blk_fwd_samps && param_fwd < _next_blk_fwd_n) _next_blk_fwd_n = param_fwd;
		}
		_params_state = PARAMS_NONINTERP;
	}

	// find src position of fft blk
	long src_fft_abs = _blk_samp_abs + _next_blk_fwd_n;

	if(_params_state == PARAMS_NONINTERP) {
		// see if we can interpolate
		if(_params.setOutPos(src_fft_abs))
			_params_state = PARAMS_INTERP_OK;
	}

	// convert params
	//
	_delay_n = getDelaySamps(_params.getInterp(BlkFxParam::DELAY));
	_plan = BlkFxParam::getPlan(_params.getInterp(BlkFxParam::FFT_LEN));

	// calculate derived variables
	//

	// find output position of the current blk using current delay
	_dst_fft_abs = src_fft_abs + _delay_n;

	// make sure dest position isn't behind current sample position
	long behind_n = _curr_samp_abs-_dst_fft_abs;
	if(behind_n > 0) {
		_next_blk_fwd_n += behind_n;
		_dst_fft_abs = _curr_samp_abs;
	}

	// number of samples to do fft blk
	_freq_fft_n = GET_FFT_BLK_SZ(_plan);
}

//-------------------------------------------------------------------------------------------------
inline void DtBlkFx::doInBuf(float** in_buf, long buf_n)
// internal method
// copy input data into the pre-fft buffer
// we may copy data beyond the end of our normal wrap position because if we do a block wraps
// over the buffer end then we do a copy (elsewhere) to from the start of the buffer anyway
// (so that the fft can get a contiguous block)
{
	int i;

	// offset into buffer at which to write data
	long t = _x0_i+_x0_n; 

	// wrap to the start of the buffer if need be
	if(t >= _x0_sz) t -= _x0_sz;

	// check if copy will pass _x0_sz
	else if(t+buf_n > _x0_sz) {
		// copy data that will overflow past _x0_sz to the start of the buffer
		long skip = _x0_sz - t;
		long n = buf_n - skip;
		for(i = 0; i < AUDIO_CHANNELS; i++)
			memcpy(&(_chan[i].x0[0]), in_buf[i]+skip, n*sizeof(float));

		// number of samples past _x0_sz
		_x0_n_past_end = n;
	}

	// copy all data into buffer (note that this may be past _x0_sz)
	for(i = 0; i < AUDIO_CHANNELS; i++)
		memcpy(&(_chan[i].x0[t]), in_buf[i], buf_n*sizeof(float));

	_x0_n += buf_n;
}

//-------------------------------------------------------------------------------------------------
inline void DtBlkFx::_process(float** in_buf, long buf_n)
// internal method
{
	//DBG("_process" << VAR(buf_n));
		
	pollUpdate(/*force*/false);

	doInBuf(in_buf, buf_n);

	// 1 + absolute position of the final sample to output now
	long buf_end_abs = _curr_samp_abs+buf_n;

	// process fft-blks in a loop
	while(_x0_n) {
		paramsChk();

		// determine whether we are forced to output a blk because of the delay or
		// we've run out of input buffer
		bool force_out = (buf_end_abs-_dst_fft_abs > 0 || _x0_n>=_x0_force_out_sz);

		// true if there's enough data in x0 to do the fft
		bool enough_data = (_x0_n >= _next_blk_fwd_n+_freq_fft_n);

		// stop looping if we're not forced to output and either the params aren't ready
		// or we don't have enough data for a blk
		if(!force_out && (_params_state != PARAMS_INTERP_OK || !enough_data)) break;

		// move forward in the input buffer
		_blk_samp_abs += _next_blk_fwd_n;
		if(_x0_n >= _next_blk_fwd_n) {
			// normal case 
			_x0_i += _next_blk_fwd_n;
			_x0_n -= _next_blk_fwd_n;

			// see whether we need to wrap to the start of the buffer (assume fwd_n is < _x0_sz)
			if(_x0_i >= _x0_sz) {
				_x0_i -= _x0_sz;
				_x0_n_past_end = 0; // no samples have been copied past the end of the buffer
			}
		}
		else {
			// special case: should never happen - all data in _x0 is used up, reset
			_x0_i = 0;
			_x0_n = 0;
			_x0_n_past_end = 0;
		}

		// process the data blk
		
		// check whether there's enough time domain data
		_time_fft_n = _freq_fft_n;
		if(!enough_data) {
			// delay is too short for requested blk sz, reduce fft length
			_plan = reducePlan(_plan, _x0_n);
			_freq_fft_n = GET_FFT_BLK_SZ(_plan);

			// zero-pad input data
			long zero_n = _freq_fft_n-_x0_n;
			if(zero_n > 0) {
				for(int i =0 ; i < AUDIO_CHANNELS; i++) {
					float* x0 = &(_chan[i].x0[_x0_i+_x0_n]);
					float* x0_end = x0 + zero_n;
					while(x0 != x0_end) *x0++ = 0.0f; 
				}
			}

			// reduce time domain fft len to match the amount of data we have
			_time_fft_n = _x0_n;
		}

		// make a contiguous blk if the current blk is split at the end of the FIFO
		long x0_sz_ext = _x0_sz+_x0_n_past_end; // 
		long split_n = (_x0_i+_time_fft_n)-x0_sz_ext;
		if(split_n > 0) {
			for(int i = 0; i < AUDIO_CHANNELS; i++)
				memcpy(&(_chan[i].x0[x0_sz_ext]), &(_chan[i].x0[_x0_n_past_end]), split_n*sizeof(float));
			_x0_n_past_end += split_n;
		}

		// do our processing
		prepOut();
		procBlk();
		nextBlk();
	}

	// zero fill x3 if it has no data to output for this blk - this will happen before the first
	// fft blk has been processed (although x3 will already be cleared) or if the delay is 
	// increased in some cases
	//
	long zero_n = buf_end_abs-_x3_end_abs;
	if(zero_n > 0 && !_x3_is_clear) {
		int i;
		long zero_o = _x3_end_abs-_curr_samp_abs;
		for(i = 0; i < AUDIO_CHANNELS; i++)
			wrapProcess(PZero(), _chan[i].x3, _x3_o+zero_o, zero_n);

		// fade out previous data (if there's any in this blk)
		long fade_n = 16;
		long fade_o = zero_o-fade_n;
		if(fade_o < 0) {
			fade_n = fade_o;
			fade_o = 0;
		}
		if(fade_n > 0) {
			for(i = 0; i < AUDIO_CHANNELS; i++)
				wrapProcess(PFadeOut(fade_n), _chan[i].x3, _x3_o+fade_o, fade_n);
		}

		_x3_end_abs = buf_end_abs;
	}

	// update absolute sample position
	_curr_samp_abs += buf_n;
}



//-------------------------------------------------------------------------------------------------
void DtBlkFx::process (float** inputs, float** outputs, long samps)
// virtual, override AudioEffect
// called by vst to process data from "inputs" and add to "outputs"
{
	_process(inputs, samps);
	for(int i = 0; i < AUDIO_CHANNELS; i++)
		wrapProcess(PAddOut(outputs[i]), _chan[i].x3, _x3_o, samps);
	_x3_o = wrap(samps+_x3_o, _chan[0].x3);
}

//-------------------------------------------------------------------------------------------------
void DtBlkFx::processReplacing (float** inputs, float** outputs, long samps)
// virtual, override AudioEffect
// called by vst to process data from "inputs" and replace data in "outputs"
{
	_process(inputs, samps);
	for(int i = 0; i < AUDIO_CHANNELS; i++)
		wrapProcess(PCopyOut(outputs[i]), _chan[i].x3, _x3_o, samps);
	_x3_o = wrap(samps+_x3_o, _chan[0].x3);
}




