/**************************************************************************************************
Copyright (c) 2005-2006 Darrell Tam
email: darrell.barrell@gmail.com

Implementation of FFTFx classes - these are the actual FFT effects called by DtBlkFx

There is one FFTFx class instance per effect type

In general the call sequence for processing data is as follows:
1. DtBLkFx has buffered a blk of data, performed FFT
2. DtBlkFx calls GetFFTFx function to retrieve the specific effect instance derived from FFTFx class
3. DtBlkFx calls the "process" method on instance from (2)
4. Process method instantiates an Effect processing class (on the stack) and calls the "Run" template function
   with the appropriate Effect
5. The run template (usually) instantiates a SplitMaskProcess (which handles processing the range
   between freqA & freqB or excluding the range between freqA & freqB)
6. The run template may instantiate another MaskProcess if the previous effect is say a harmonic mask
7. The run method from the SplitMaskProcess (or other MaskProcess) is called
8. SplitMaskProcess (or other MaskProcess) calls run method of effect class passed into Run template
   function 0 or more times for each continuous piece of the spectrum that needs processing
   
The key things are:
   Heavy use of templates to try to force as much compile time optimization as possible.

   There are 2 levels of classes for each effect: the interface based on FFTFx where there is one instance
   per effect and the actual effect processing that is derived from the internal class "ProcessBase".

   Masks and effects are both based on "ProcessBase" and calls flow downwards via the "run" method of
   "ProcessBase" that they override.

   The ProcessBase effects are instantiated on the stack, so that means you can't store any data between
   blocks in them

History
	Date       Version    Programmer         Comments
	16/2/03    1.0        Darrell Tam		Created



This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

***************************************************************************************************/

#define DBG_FILE_NAME "c:\\fftfx.txt"
#include "Debug.h"

#include <dt/SinCosTable.h>

#include "process_loops.h"
#include "DtBlkFx.hpp"
#include "FFTFx.h"
#include "HarmData.h"

using namespace std;

// constants
enum {
	AUDIO_CHANNELS = BlkFxParam::AUDIO_CHANNELS

};

static SinCosTable<cplxf, /*bits*/12> sincos_table;
long g_rand_i = 1;

//*************************************************************************************************
class PhaseCorrect
// find phase correction for bin shifting operations
{
public:
	// absolute sample position multiplied by sincos table length
	long long _mult;

	// right shift to divide by fft length
	int _shift;

	// most recent correction
	cplxf _curr;
	cplxf& operator () () { return _curr; }

	cplxf operator () (int bin_shift)
	{
		_curr = sincos_table[(_mult*bin_shift)>>_shift];
		return _curr;
	}

	// MUST call init before use
	void init(DtBlkFx* b)
	{
		_mult = (long long)b->_blk_samp_abs * sincos_table.LEN;
		_shift = b->_plan+FFT_LOG2_BLK_SZ_0;
	}
};

//*************************************************************************************************
class ProcessBase
// base class for effects
{
public:
	bool reverse() { return false; }

	void prepare(ProcessBase* parent) {}
	void run(long b0, long b1) {}
	void done() {}

	float pwrChg(int ch) { return 0.0f; }
};

//*************************************************************************************************
class AmpProcess : public ProcessBase
//
// default process multiplies data by amplitude
//
{
public:
	Ptr<DtBlkFx> _b;

	// amplify to apply
	float _amp;

	// power change due to this effect
	float _pwr_chg[AUDIO_CHANNELS];

	void init(DtBlkFx* b, float amp)
	{
		_b = b;
		_amp = amp;
		for(int ch = 0; ch < AUDIO_CHANNELS; ch++) _pwr_chg[ch] = 0.0f;

	}

	AmpProcess(DtBlkFx* b, float amp) { init(b, amp); }
	AmpProcess(DtBlkFx* b) { init(b, b->currFxParams().amp); }

	// apply scaling to bins b0 to b1
	void run(long b0, long b1)
	{
		float in_pwr_scale = (_amp*_amp-1.0f);
		for(int ch = 0; ch < AUDIO_CHANNELS; ch++) {
			float in_pwr = 0.0f;
			for(pcplxf_range x(_b->FFTdata(ch), b0, b1+1); !x.empty(); x++) {
				in_pwr += norm(*x);
				*x = (*x) * _amp;
			}
			_pwr_chg[ch] += in_pwr*in_pwr_scale;
		}
	}

	// 
	float pwrChg(int ch) { return _pwr_chg[ch]; }
};

//*************************************************************************************************
inline float/*pwr*/GetPwr(pcplxf_range x/*must be fwd*/)
{
	float pwr = 0.0f;
	for(; !x.empty(); x++) pwr += norm(*x);
	return pwr;
}
inline float/*pwr*/GetPwr(pcplxf x, long b0, long b1 /*b0 <= b1*/)
	{ return GetPwr(pcplxf_range(x, b0, b1+1)); }

inline float/*amp*/MatchPwr(float amp, double target_pwr, double curr_pwr)
{
	// find scaling to match output pwr to input pwr
	double s = amp;
	if(target_pwr <= 0) return 0;
	if(curr_pwr > 0) s *= sqrt(target_pwr/curr_pwr);

	// limit the scaling
	if(s > 1e20) return 1.0f;
	if(s < 1e-30) return 0.0f;
	return (float)s;

}
inline void MatchPwr(float amp, double target_pwr, double curr_pwr, pcplxf_range x /*must be fwd*/)
{
	amp = MatchPwr(amp, target_pwr, curr_pwr);
	for(; !x.empty(); x++)
		*x = (*x) * amp;
}

//*************************************************************************************************
template <class T> class MaskProcessBase : public ProcessBase
// base class for mask processing
{
public:
	Ptr<T> _process;

	// default to calling through to next processing
	bool reverse() { return _process->reverse(); }

	void prepare(ProcessBase* parent) { _process->prepare(this); }

	void done() { _process->done(); }

	float pwrChg(int ch) { return _process->pwrChg(ch); }
};

//*************************************************************************************************
struct HarmParam
// harmonic parameters derived from a frequency and fx value
{
	// multiplier being used on fundamental
	float _multiply;

	// centre of harm-0 (bin)
	float _offs;

	// width of harmonics (bins)
	long _width;
	long _half_width;

	// spacing of harmonics (bins)
	float _spacing;

	// maximum bin (set to limit the number of harmonics processed)
	long _max_bin;

	// minimum bin (edge of lowest harmonic to process)
	long _min_bin;

	// original fundamental centre frequency bin supplied
	float _freq;

	HarmParam() { _multiply = 1.0f; }

	// initialize
	void init(
		float/*0..1*/fx_val,	//
		float freq				// bin pos of fundamental freq
	)
	// initialize, expect _multiply to already be set
	{
		// note, fundamental freq is considered to be numbered harmonic-1 for odd & even
		float f = freq*_multiply;

		if(f < 1) {
			// disable effect if frequency is invalid
			_multiply = 1;
			_freq = 1;
			_offs = 1;
			_spacing = 1;
			_width = 0;
			_half_width = 0;
			_max_bin = 0;
			_min_bin = 1;
			return;
		}

		_freq = freq;

		SplitParam<4> h(fx_val);

		switch(h.i_part) {
		case 0: _offs = f; _spacing = f; break; // all harmonics
		case 1: _offs = f; _spacing = f*2; break; // odd harmonics
		case 2: _offs = f*2; _spacing = f*2; break; // even harmonics
		case 3: _offs = f*0.5f; _spacing = f; break; // in between harmonics
		}

		// don't let harmonics get too small
		if(_spacing < 4) _spacing = 4;

		float fwidth = limit_range(_spacing*h.f_part, 0.0f, _spacing);
		_half_width = (long)fwidth/2;
		_width = _half_width*2;

		// don't do more than 100 harmonics
		_max_bin = (long)(_spacing*200);

		// don't go below fundamental (i.e. don't treat 0Hz as a harmonic)
		_min_bin = (long)_offs-_half_width;
	}

};

//*************************************************************************************************
template <class T> class HarmMaskProcess : public MaskProcessBase<T>, public HarmParam
{
public:

	// current centre bin
	float _curr_cent;

	// current harmonic (0 = fundamental*multiplier)
	float _curr_harm;

	void prepare(ProcessBase *parent) { _process->prepare(this); }

	HarmMaskProcess(T& process, float/*0..1*/fx_val, float/*bin pos*/harm0)
	{
		/*MaskProcessBase::*/_process = &process;
		init(fx_val, harm0);
	}

	// round a frequency down to harmonic centre
	void roundHarm(float bin) {
		_curr_harm = floorf((bin-_offs)/_spacing);
		_curr_cent = _offs+_spacing*_curr_harm;
	}

	// run the process
	void run(long b0, long b1)
	{
		if(b0 < _min_bin) b0 = _min_bin;
		if(b1 > _max_bin) b1 = _max_bin;
		if(b0 > b1) return;

		bool looping = true;
		if(reverse()) {
			// process spectrum in reverse
			roundHarm(b1+_spacing);

			// max bin for harmonic
			long p = b1;
			while(looping) {
				long v1 = (long)(_curr_cent+0.5f)+_half_width;
				long v0 = v1-_width;
				// don't go outside our range
				if(v0 <= b0) {
					v0 = b0;
					looping = false;
				}
				if(v1 > p) v1 = p;
				// process if range is good
				if(v1 >= v0) {
					_process->run(v0, v1);
					p = v0-1;
				}
				_curr_cent -= _spacing;
				_curr_harm--;
			}
		}
		else {
			// process spectrum fwd
			roundHarm((float)b0);

			// min bin for harmonic
			long p = b0;
			while(looping) {
				long v0 = (long)(_curr_cent+0.5f)-_half_width;
				long v1 = v0+_width;
				// don't go outside range
				if(v1 >= b1) {
					v1 = b1;
					looping = false;
				}
				if(v0 < p) v0 = p;
				// process if range is good
				if(v1 >= v0) {
					_process->run(v0, v1);
					p = v1+1;
				}
				_curr_cent += _spacing;
				_curr_harm++;
			}
		}
	}
protected:
	// if you don't construct with anything, make sure that you:
	// set _process, _multiply
	// call init()
	HarmMaskProcess() {}


};
//-------------------------------------------------------------------------------------------------
template <int N_CHANS> struct PeakFindT
// find peak bin from effect between b0 & b1 inclusive
{
	float max_pwr;
	long max_bin;

	PeakFindT() {}

	PeakFindT(Array<pcplxf, N_CHANS> ch_dat, long b0, long b1) { operator() (ch_dat, b0, b1); }

	void operator () (Array<pcplxf, N_CHANS> ch_dat, long b0, long b1 /* b1 >= b0 */)
	{
		max_pwr = 0.0f;

		int ch;
		float* ch0_start_r = ch_dat[0].r;
		float* ch0_end_r = ch_dat[0].r + b1 + 1;

		for(ch = 0; ch < N_CHANS; ch++) ch_dat[ch] += b0;

		// arbitrary value if no peak found
		float* max_p = ch_dat[0].r; 
		while(ch_dat[0].r != ch0_end_r) {
			float t = norm(*ch_dat[0]);
			for(ch = 1; ch < N_CHANS; ch++) t += norm(*ch_dat[ch]);
			if(t > max_pwr) {
				max_pwr = t;
				max_p = ch_dat[0].r;
			}
			for(ch = 0; ch < N_CHANS; ch++) ch_dat[ch]++;
		}
		max_bin = max_p - ch0_start_r;
	}

	// return peak bin position
	operator long() { return max_bin; }
};

//-------------------------------------------------------------------------------------------------
struct PeakFind1Ch : public PeakFindT<1>
// find peak bin for a single channel between b0 & b1 inclusive
{
	PeakFind1Ch() {}

	PeakFind1Ch(pcplxf dat, long b0, long b1) { operator() (dat, b0, b1); }

	void operator () (pcplxf dat, long b0, long b1)
	{
		Array<pcplxf, 1> ch_dat;
		ch_dat[0] = dat;
		PeakFindT<1>::operator () (ch_dat, b0, b1);
	}
};

//-------------------------------------------------------------------------------------------------
struct PeakFind : public PeakFindT<AUDIO_CHANNELS>
// find peak bin from effect between b0 & b1 inclusive
{
	PeakFind() {}

	PeakFind(DtBlkFx* b, long b0, long b1) { operator() (b, b0, b1); }

	void operator () (DtBlkFx* b, long b0, long b1)
	{
		int ch;
		Array<pcplxf, AUDIO_CHANNELS> ch_dat;
		for(ch = 0; ch < AUDIO_CHANNELS; ch++) ch_dat[ch] = b->FFTdata(ch);
		PeakFindT<AUDIO_CHANNELS>::operator () (ch_dat, b0, b1);
	}
};

//-------------------------------------------------------------------------------------------------
template <class T> class AutoHarmMaskProcess : public HarmMaskProcess<T>
//
// auto harm mask aligns fundamental freq with peak within freq a/b within params
//
{
public:

	AutoHarmMaskProcess(T& process, DtBlkFx* b, FxParams& params, float fx_val, float multiply)
	{
		_process = &process;
		_multiply = multiply;

		Array<long, 2> bin = params.bin;
		if(bin[0] > bin[1]) swap(bin[0], bin[1]);

		// peak find on all channels added together
		long peak_bin = PeakFind(b, bin[0], bin[1]);
		/*HarmMaskProcess::*/init(fx_val, (float)peak_bin);
	}
};

//-------------------------------------------------------------------------------------------------
template <class T> class SplitMaskProcess : public MaskProcessBase<T>
{
public:
	// bins from the params
	long _param_bin[2];

	SplitMaskProcess(T& process, FxParams& params)
	{
		/*MaskProcessBase::*/_process = &process;
		_param_bin[0] = params.bin[0];
		_param_bin[1] = params.bin[1];
	}

	void run(long b0, long b1)
	{
		if(_param_bin[0] <= _param_bin[1]) {
			// freqA < freqB : normal process inside region
			long v0 = max(b0, _param_bin[0]);
			long v1 = min(b1, _param_bin[1]);
			if(v1 >= v0) _process->run(v0, v1);
		}
		else {
			// freqA > freqB : process outside region
			long v1 = min(b1, _param_bin[1]);
			long v0 = max(b0, _param_bin[0]);
			if(reverse()) {
				if(v1 >= b0) _process->run(b0, v1);
				if(b1 >= v0) _process->run(v0, b1);
			}
			else {
				if(b1 >= v0) _process->run(v0, b1);
				if(v1 >= b0) _process->run(b0, v1);
			}
		}
	}
};

//-------------------------------------------------------------------------------------------------
template <class T> class InvMaskProcess_ : public MaskProcessBase<T>
//
// invert ranges processed
//
{
public:
	// inverted range is limited to this range
	long _min_bin, _max_bin;

	long _prev_bin;

	void init(long min_bin, long max_bin)
	{
		_min_bin = min_bin;
		_max_bin = max_bin;
		_prev_bin = reverse()?_max_bin:_min_bin;
	}

	// do the processing
	void run(long b0, long b1)
	{
		if(reverse()) {
			long v0 = max(b1+1, _min_bin);
			if(v0 <= prev_bin) {
				_process->run(v0, _prev_bin);
				_prev_bin = b0-1;
			}
		}
		else {
			long v1 = min(b0-1, _max_bin);
			if(prev_bin <= v1) {
				_process->run(_prev_bin, v1);
				_prev_bin = b1+1;
			}
		}
	}
	void runFinal()
	{
		if(reverse()) {
			if(_min_bin <= _prev_bin) _process->run(_min_bin, _prev_bin);
		}
		else {
			if(_prev_bin <= _max_bin) _process->run(_prev_bin, _max_bin);
		}
	}
};

//-------------------------------------------------------------------------------------------------
template <class T, class M> class InvMaskProcess : public MaskProcessBase<M>
//
// Processing chain is:
//     InvMaskProcess<T, M<InvMaskProcess_<T> > to M< InvMaskProcess_<T> > to InvMaskProcess_ <T> to T
//
{
	InvMaskProcess_<T> _inv_mask;

public:
	InvMaskProcess(T& process, M& mask)
	{
		_inv_mask._process = process;
		/*MaskProcessBase::*/_process = mask;
	}

	void run(long b0, long b1)
	{
		_inv_mask.init(b0, b1);
		_process->run(b0, b1);
		_inv_mask.runFinal();
	}
};

//-------------------------------------------------------------------------------------------------
template <class T> class ThreshMaskProcess : public MaskProcessBase<T>
//
// ranges correspond to bins above a certain threshold
{
public:
	DtBlkFx* _b;

	float _thresh;

	// width either side of a threshold hit to include
	long _width_bins;

	// width_bins*2 + 1
	long _width2_bins;

	// true=select areas below threshold, false=select above threshold
	bool _select_below;


	ThreshMaskProcess(
		T& process,
		DtBlkFx* b,					//
		float thresh_param			// param to set width or threshold
	)
	{
		/*MaskProcessBase::*/_process = &process;

		_b = b;

		//_width_bins = (long)(width_frac*(float)b->_freq_fft_n);
		_width_bins = 1;
		_width2_bins = _width_bins*2+1;

		SplitParam<2> split(thresh_param);
		_select_below = split.i_part?true:false;

		// find total pwr in all channels
		float total_pwr = 0.0f;
		for(int ch = 0; ch < AUDIO_CHANNELS; ch++) total_pwr += b->_chan[ch].total_out_pwr;

		// arbitrary threshold function, -40 .. +40 dB above average bin power
		_thresh = powf(10.0f, (split.f_part-0.5f)*4.0f) * total_pwr / (float)(b->_freq_fft_n);
	}

	bool/*found*/testThresh(float x, bool select_below)
	// select_below=true: return true if "x" is less than the theshold
	// select_below=false: return true if "x" is above the threshold
	{
		if(select_below) { if(x < _thresh) return true; }
		else { if(x > _thresh) return true; }
		return false;
	}

	bool/*peak found*/findThreshBrk(
		pcplxf_range /*in-out*/ dat[],	// only dat[0] is tested for empty()
		bool select_below,
		int step						// 1=fwd, -1=rev
	)
	{
		int ch;
		while(!dat[0].empty()) {
			float t = 0.0f;
			for(ch = 0; ch < AUDIO_CHANNELS; ch++) t += norm(*(dat[ch]));
			if(testThresh(t, select_below)) return true;
			for(ch = 0; ch < AUDIO_CHANNELS; ch++) dat[ch] += step;
		}
		return false;
	}


	void run(long b0, long b1, bool select_below)
	{
		long v0 = -1, v1 = -1;
		int ch;
		float* dat_ch0_r = _b->FFTdata(/*channel*/0).r;
		pcplxf_range dat[AUDIO_CHANNELS];

		// find points breaking the threshold and include "width" data either side
		if(reverse()) {
			// find first bin to break threshold
			dat[0] = pcplxf_range(_b->FFTdata(/*channel*/0), b1, b0-1);
			for(ch = 1; ch < AUDIO_CHANNELS; ch++) dat[ch] = _b->FFTdata(ch)+b1;

			if(findThreshBrk(dat, select_below, /*dir*/-1))
				v1 = min(dat[0].r-dat_ch0_r+_width_bins, b1);

			float* prev_dat_ch0_r = dat[0].r;

			// loop finding points breaking threshold
			while(!dat[0].empty()) {
				prev_dat_ch0_r = dat[0].r;
				for(ch = 0; ch < AUDIO_CHANNELS; ch++) dat[ch]--;
				if(!findThreshBrk(dat, select_below, /*dir*/-1)) break;

				// distance from previous theshold breaker
				long dist = prev_dat_ch0_r - dat[0].r;
				if(dist > _width2_bins) {
					// run prev range
					v0 = max(prev_dat_ch0_r-dat_ch0_r-_width_bins, b0);
					if(v0 <= v1) _process->run(v0, v1);

					// set start position for this rng
					v1 = dat[0].r-dat_ch0_r+_width_bins;
				}
			}

			// run final rng
			if(v1 >= 0) {
				v0 = max(prev_dat_ch0_r-dat_ch0_r-_width_bins, b0);
				if(v0 <= v1) _process->run(v0, v1);
			}
		}
		else {
			// find initial threshold break
			dat[0] = pcplxf_range(_b->FFTdata(/*channel*/0), b0, b1+1);
			for(ch = 1; ch < AUDIO_CHANNELS; ch++) dat[ch] = _b->FFTdata(ch)+b0;
			if(findThreshBrk(dat, select_below, /*dir*/1))
				v0 = max(dat[0].r-dat_ch0_r-_width_bins, b0);

			float* prev_dat_ch0_r = dat[0].r;

			// loop finding points breaking threshold
			while(!dat[0].empty()) {
				prev_dat_ch0_r = dat[0].r;
				for(ch = 0; ch < AUDIO_CHANNELS; ch++) dat[ch]++;
				if(!findThreshBrk(dat, select_below, /*dir*/1)) break;

				// distance from previous theshold breaker
				long dist = dat[0].r - prev_dat_ch0_r;
				if(dist > _width2_bins) {
					// run prev range
					v1 = min(prev_dat_ch0_r-dat_ch0_r+_width_bins, b1);
					if(v0 <= v1) _process->run(v0, v1);

					// set start position for this rng
					v0 = dat[0].r-dat_ch0_r-_width_bins;
				}
			}

			// run final rng
			if(v0 >= 0) {
				v1 = min(prev_dat_ch0_r-dat_ch0_r+_width_bins, b1);
				if(v0 <= v1) _process->run(v0, v1);
			}
		}
	}

	void run(long b0, long b1)
	{
		// force inlining by passing "select below" as constant
		if(_select_below) run(b0, b1, /*select below=*/true);
		else run(b0, b1, /*select below=*/false);
	}
};

//*************************************************************************************************
int HarmDispVal(char* /*out*/text, float/*0..1*/val)
{
	const char* name[] = {"both", "odd", "even", "between"};
	SplitParam<4> h(val);
	return sprintf(text, "%.0f%% %s", h.f_part*100.0f, name[h.i_part]);
}

//-------------------------------------------------------------------------------------------------
class HarmMaskFx : public FFTFx
// harmonic mask effect does no processing - it provides a freq mask for the next effect
{
public:
	HarmMaskFx(const char* name, bool freq_b_param_used) : FFTFx(name)
	{
		_params_used[BlkFxParam::FX_FREQ_B] = freq_b_param_used;
		_params_used[BlkFxParam::FX_AMP] = false;
	}

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return HarmDispVal(text, val); }
};
HarmMaskFx g_harm_mask("HarmMask", /*freq_b_param_used*/false);
HarmMaskFx g_auto_harm_mask("AutoHarmMask", /*freq_b_param_used*/true);
HarmMaskFx g_auto_subharm1_mask("ASubH1Mask", /*freq_b_param_used*/true);
HarmMaskFx g_auto_subharm2_mask("ASubH2Mask", /*freq_b_param_used*/true);
HarmMaskFx g_auto_subharm3_mask("ASubH3Mask", /*freq_b_param_used*/true);

//*************************************************************************************************
int ThreshDispVal(char* /*out*/text, float/*0..1*/val)
{
	const char* type[2] = { "above", "below" };
	SplitParam<2> split(val);
	return sprintf(text, "%s %2.0f%%", type[split.i_part], split.f_part*100.0f);
}

//-------------------------------------------------------------------------------------------------
class ThreshMaskFx : public FFTFx
{
public:
	ThreshMaskFx() : FFTFx("ThreshMask")
	{
		_params_used[BlkFxParam::FX_AMP] = false;
		_params_used[BlkFxParam::FX_FREQ_A] = false;
		_params_used[BlkFxParam::FX_FREQ_B] = false;
	}

	virtual void process(DtBlkFx* b) {}

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return ThreshDispVal(text, val); }

} g_thresh_mask;


//-------------------------------------------------------------------------------------------------
template<class T> inline void run_(DtBlkFx* b, T& process, long b0, long b1)
// run "process" over bins b0->b1 & adjust pwr
{
	process.prepare(/*parent*/NULL);
	if(b0 <= b1) process.run(b0, b1);
	process.done();
	for(int ch = 0; ch < AUDIO_CHANNELS; ch++)
		b->_chan[ch].total_out_pwr += process.pwrChg(ch);
}

//-------------------------------------------------------------------------------------------------
template<class T> inline void run_(DtBlkFx* b, T& process, bool use_split_mask, long b0, long b1)
{
	if(use_split_mask) run_(b, SplitMaskProcess<T>(process, b->currFxParams()), b0, b1);
	else run_(b, process, b0, b1);
}

//-------------------------------------------------------------------------------------------------
bool getHarmMaskFreqMult(FFTFx* prev_fx, float& /*out*/ mult)
// return freq multiplier if the previous effect is a harmonic mask
{
	if(prev_fx == &g_auto_harm_mask) mult = 1.0f;
	else if(prev_fx == &g_auto_subharm1_mask) mult = 0.5f;
	else if(prev_fx == &g_auto_subharm2_mask) mult = 1.0f/3.0f;
	else if(prev_fx == &g_auto_subharm3_mask) mult = 0.25f;
	else return false;
	return true;
}

//-------------------------------------------------------------------------------------------------
template<class T> inline bool/*ran mask*/runHarmMask(
	DtBlkFx* b,
	T& end_process,
	bool use_split_mask,
	long b0,
	long b1
)
// check whether the effect needs to run with a harm mask based on the previous effect
// if so, run the effect with the harmonic mask
{
	// check what the prev params were

	FxParams* prev_params = b->prevFxParams();;
	if(!prev_params) return false;

	FFTFx *prev_fx = prev_params->fft_fx;

	if(prev_fx == &g_harm_mask) {
		run_(b,
			HarmMaskProcess<T>(end_process, prev_params->val, prev_params->fbin[0]),
			use_split_mask,
			b0, b1
		);
		return true;
	}
	float mult;
	if(!getHarmMaskFreqMult(prev_fx, /*out*/ mult)) return false;

	AutoHarmMaskProcess<T> autoharm(end_process, b, *prev_params, prev_params->val, mult);
	run_(b, autoharm, use_split_mask, b0, b1);
	return true;
}

//-------------------------------------------------------------------------------------------------
template<class T> inline void run(
	DtBlkFx* b,
	T& end_process,
	bool use_split_mask,
	long b0,
	long b1
)
//
// default BlkFx process calls 
//
{
	FxParams* prev_params = b->prevFxParams();
	if(prev_params) {
		if(runHarmMask(b, end_process, use_split_mask, b0, b1)) return;

		// check for other masks
		if(prev_params->fft_fx == &g_thresh_mask) {
			run_(b,
				ThreshMaskProcess<T>(end_process, b, prev_params->val),
				use_split_mask,
				b0, b1
			);
			return;
		}
	}

	// no previous mask effect, do normal process
	run_(b, end_process, use_split_mask, b0, b1);
}

// run over entire range
template<class T> inline void run(DtBlkFx* b, T& end_process, bool use_split_mask=true)
	{ run(b, end_process, use_split_mask, /*start bin*/1, /*end bin*/b->_freq_fft_n/2-1); }


//*************************************************************************************************
class HarmFiltFx : public FFTFx
{
public:
	HarmFiltFx() : FFTFx("HarmFilt") {}

	virtual void process(DtBlkFx* b)
	{
		FxParams& params = b->currFxParams();
		AmpProcess amp(b);
		HarmMaskProcess<AmpProcess> harm(
			amp,
			params.val, // value
			min(params.fbin[0], params.fbin[1]) // centre
		);
		run(b, harm, /*split mask*/false, /*start bin*/1, /*end bin*/max(params.bin[0], params.bin[1]));
	}

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return HarmDispVal(text, val); }

} g_harm_filt_fx;



//*************************************************************************************************
// no effect does nothing
class NoFx : public FFTFx
{
public:
	NoFx() : FFTFx("NoEffect")
	{
		memset(_params_used, 0, sizeof(_params_used));
	}
} g_no_fx;

//*************************************************************************************************
// filter effect
class FilterFx : public FFTFx
{
public:
	FilterFx() : FFTFx("Filter")
	{
		_params_used[BlkFxParam::FX_VAL] = false;
	}
	virtual void process(DtBlkFx* b) { run(b, AmpProcess(b)); }

} g_filter_fx;


//*************************************************************************************************
struct ContrastProcess : public AmpProcess
// "contrast" effect performs abs(x) ^ (2*raise+1) over the spectrum
//
{
	static float scaleVal(float v)
	{
		v -= 0.5f;
		if(v >= 0.0f) return lin_interp(v*v*4.0f, 0.0f, 4.0f);
		else return lin_interp(v*v*4.0f, 0.0f, -0.5f);
	}


	// pwr to raise to
	float _raise;

	float _scale[AUDIO_CHANNELS];

	ContrastProcess(DtBlkFx* b) : AmpProcess(b)
	{
		_raise = scaleVal(b->currFxParams().val);

		// find a scale value to make the total spectrum pwr 10.0f (arbitrary)
		for(int ch = 0; ch < AUDIO_CHANNELS; ch++) {
			float out_pwr = b->_chan[ch].total_out_pwr;
			if(out_pwr > 1e-20) _scale[ch] = 10.0f/out_pwr;
			else _scale[ch] = 1.0f;
		}
	}

	void run(long b0, long b1)
	{
		for(int ch = 0; ch < AUDIO_CHANNELS; ch++) {
			// input power for this range
			float in_pwr = 0;

			// output power for this range
			float out_pwr = 0;

			float scale = _scale[ch];

			// variables from AmpProcess: _b, _pwr_chg, _amp
			if(_raise >= 0) {
				// raise positive
				for(pcplxf_range x(_b->FFTdata(ch), b0, b1+1); !x.empty(); x++) {
					float t = norm(*x); 
					in_pwr += t;
					*x = (*x) * powf(t*scale, _raise); // equivalent to abs(*x) = abs(*x)^(2*raise+1) 
					out_pwr += norm(*x);
				}
			}
			else {
				// raise negative
				for(pcplxf_range x(_b->FFTdata(ch), b0, b1+1); !x.empty(); x++) {
					float t = norm(*x); 
					in_pwr += t;
					if(t > 1e-20)
						*x = (*x) * powf(t*scale, _raise); // equivalent to abs(*x) = abs(*x)^(2*raise+1) 
					out_pwr += norm(*x);
				}
			}

			// match output pwr to input power
			MatchPwr(_amp, in_pwr, out_pwr, pcplxf_range(_b->FFTdata(ch), b0, b1+1));

			// 
			_pwr_chg[ch] += in_pwr;
		}
	}

	float pwrChg(int ch) { return _pwr_chg[ch]*(_amp*_amp-1); }
};

//-------------------------------------------------------------------------------------------------
class ContrastFx : public FFTFx
{
public:
	ContrastFx() : FFTFx("Contrast") { }

	virtual void process(DtBlkFx* b) { run(b, ContrastProcess(b)); }

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return sprintf(text, "%.0f", (val-0.5f)*200.0f); }

} g_contrast_fx;


//******************************************************************************************
struct SmearProcess : public AmpProcess
{
	float _smear;

	SmearProcess(DtBlkFx* b) : AmpProcess(b)
	{
		_smear = b->currFxParams().val;
	}

	void run(long b0, long b1)
	//
	// randomize the phase
	//
	{
		long rand_i = g_rand_i;
		for(int ch = 0; ch < AUDIO_CHANNELS; ch++) {
			for(pcplxf_range x(_b->FFTdata(ch), b0, b1+1); !x.empty(); x++) {
				/*AmpProcess::*/_pwr_chg[ch] += norm(*x);
				*x = /*AmpProcess::*/_amp * (*x) * (sincos_table[rand_i]*_smear+1.0f-_smear);
				rand_i = prbs29(rand_i);
			}
		}
		g_rand_i = rand_i;
	}

	float pwrChg(int ch) { return _pwr_chg[ch]*(_amp*_amp-1); }
};


//-------------------------------------------------------------------------------------------------
class SmearFx : public FFTFx
{
public:
	SmearFx() : FFTFx("Smear") {}

	virtual void process(DtBlkFx* b) { run(b, SmearProcess(b)); }

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return sprintf(text, "%2.0f%%", val*100.0f); }
} g_smear_fx;



//-------------------------------------------------------------------------------------------------
class ThreshFx : public FFTFx
{
public:
	ThreshFx() : FFTFx("Threshold") {}

	virtual void process(DtBlkFx* b)
	{
		FxParams& params = b->currFxParams();
		AmpProcess amp(b);
		ThreshMaskProcess<AmpProcess> thresh(amp, b, /*thresh param*/params.val);
		run(b, thresh);
	}

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return ThreshDispVal(text, val); }

} g_thresh_fx;


//*************************************************************************************************
struct ClipProcess : public AmpProcess
{

	// threshold for clipping
	float _thresh[AUDIO_CHANNELS];

	// sqrt thresh
	float _root_thresh[AUDIO_CHANNELS];

	ClipProcess(DtBlkFx* b) : AmpProcess(b)
	{
		// arbitary threshold function
		for(int ch = 0; ch < AUDIO_CHANNELS; ch++) {
			_thresh[ch] = 1.0f - b->currFxParams().val;
			_thresh[ch] *= _thresh[ch]*8.0f*b->_chan[ch].total_out_pwr/b->_freq_fft_n;
			if(_thresh[ch] < 1e-30f) _thresh[ch] = 1e-30f;

			_root_thresh[ch] = sqrtf(_thresh[ch]);
		}
	}
	void run(long b0, long b1)
	// clip bins with magnitude higher than the threshold
	{
		for(int ch = 0; ch < AUDIO_CHANNELS; ch++) {
			float in_pwr = 0;
			float out_pwr = 0;
			for(pcplxf_range x(_b->FFTdata(ch), b0, b1+1); !x.empty(); x++) {
				float t0 = norm(*x);
				in_pwr += t0;
				if(t0 >= _thresh[ch]) {
					float t1 = _root_thresh[ch] / sqrtf(t0);
					*x = (*x) * t1;
					out_pwr += _thresh[ch];
				}
				else out_pwr += t0;
			}
			MatchPwr(_amp, in_pwr, out_pwr, pcplxf_range(_b->FFTdata(ch), b0, b1+1));
			/*AmpProcess::*/_pwr_chg[ch] += in_pwr;
		}
	}
	float pwrChg(int ch) { return _pwr_chg[ch]*(_amp*_amp-1); }
};

//-------------------------------------------------------------------------------------------------
class ClipFx : public FFTFx
{
public:
	ClipFx() : FFTFx("Clip") {}

	virtual void process(DtBlkFx* b) { run(b, ClipProcess(b)); }

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return sprintf(text, "%2.0f%%", val*100.0f); }

} g_clip_fx;


//-------------------------------------------------------------------------------------------------
static float/*octaves shift*/GetOctShift(float/*0..1*/v)
	{ return lin_interp(v, -3.0f, 3.0f); }

//*************************************************************************************************
static float/*Hz*/GetFreqShift(float v/*0..1*/)
//
// Convert freq shift param to Hz
//
// return frequency as a fraction between -22050 to 22050 Hz
// maps the linear input "v" to an arbitary linear+exponential function
// (linear transitioning into exponential)
//
{
	const float lin = 2048.0f/32768.0f;
	const int exp_raise = 8;
	const float exp_scale = (1.0f-lin)/(1<<exp_raise);
	v = (v-0.5f)*2;

	bool pos = true;
	if(v < 0) {
		v = -v;
		pos = false;
	}
	float r = (powf(2, v*exp_raise)-1.0f)*exp_scale + v*lin;
	return (pos ? r:-r)*22050.0f;
}
//-------------------------------------------------------------------------------------------------
template <class T/*dummy class type*/> class ShiftProcess : public AmpProcess
//
{
public:
	// channel being processed
	int _ch;

	// max bin
	long _max_bin;

	// phase correction
	PhaseCorrect _phase;

	// true=constant shift, false=scale shift
	bool _do_const_shift;

	// bin range meaning: true=destination, false=source
	bool _dest_mode;

	// bin shift distance
	long _const_bin_shift;

	// variables only used for scaled freq shift (_do_const_shift=false)

	// inverse scale mult used for dest_mode only
	float _inv_freq_scale;

	// centre freq multiplier to do scaled shift
	float _shift_mult;

	float _bin_to_hz, _hz_to_bin;

	// lowest bin that we will do type a shifting
	long _type_a_low_bin;
	long _type_b_width;
	

	// amplitude to apply to src
	float _mix_amp;
	
	// power of input prior to effect
	float _src_pwr[AUDIO_CHANNELS];

	// direction to run processing
	bool _reverse;
	bool reverse() { return _reverse; }

	//
	ShiftProcess(DtBlkFx* b, bool dest_mode) : AmpProcess(b)
	{
		_phase.init(b);

		for(int i = 0; i < AUDIO_CHANNELS; i++) _src_pwr[i] = 0.0f;

		_dest_mode = dest_mode;

		// if amp is less than 1 then we're in mix-mode
		if(_amp < 1.0f) _mix_amp = 1.0f - _amp;
		else _mix_amp = 0.0f;

		long fft_len = b->_freq_fft_n;
		_max_bin = fft_len/2-1;

		// split param into 2
		SplitParam<2> split(b->currFxParams().val);

		// process in reverse when shifting up in frequency

		if(split.i_part) {
			// constant frequency shift
			_do_const_shift = true;
			_const_bin_shift = (long)(
				GetFreqShift(split.f_part)/b->getSampleRate()*(float)fft_len + 0.5f
			);
			_phase() = _phase(_const_bin_shift) * AmpProcess::_amp;

			_reverse = (_const_bin_shift > 0);
		}
		else {
			// scaled frequency shift
			_do_const_shift = false;
			float freq_scale = powf(2.0f, GetOctShift(split.f_part));
			_shift_mult = freq_scale-1.0f;
			_reverse = (freq_scale > 1.0f);
			if(_dest_mode) _inv_freq_scale = 1.0f/freq_scale;

			_bin_to_hz = _b->getSampleRate()/(float)fft_len;
			_hz_to_bin = (float)fft_len/_b->getSampleRate();

			// arbitrary sizes
			_type_a_low_bin = fft_len/130;
			_type_b_width = max(fft_len/700, 1L);
		}
	}

	// run method is dependent on "T"
	void run(pcplxf dst, pcplxf src);

	void run(long dst_bin, long src_bin, long n_bins)
	{
		if(src_bin < 1) {
			dst_bin += (1-src_bin);
			n_bins -= (1-src_bin);
			src_bin = 1;
		}
		if(src_bin+n_bins > _max_bin) n_bins = _max_bin - src_bin;

		if(dst_bin < 1)  {
			src_bin += (1-dst_bin);
			n_bins -= (1-dst_bin);
			dst_bin = 1;
		}
		if(dst_bin+n_bins > _max_bin) n_bins = _max_bin - dst_bin;

		if(n_bins < 1) return;

		if(_reverse) {
			// process spectrum in reverse (shift up)
			for(_ch = 0; _ch < AUDIO_CHANNELS; _ch++) {
				pcplxf xd(_b->FFTdata(_ch), dst_bin+n_bins);
				for(pcplxf_range xs(_b->FFTdata(_ch), src_bin+n_bins-1, src_bin-1); !xs.empty(); xs--, xd--) {
					_pwr_chg[_ch] -= norm(*xd);
					run(xd, xs);
					_pwr_chg[_ch] += norm(*xd);
				}
			}
		}
		else {
			// process spectrum forwards (shift down)
			for(_ch = 0; _ch < AUDIO_CHANNELS; _ch++) {
				pcplxf xd(_b->FFTdata(_ch), dst_bin);
				for(pcplxf_range xs(_b->FFTdata(_ch), src_bin, src_bin+n_bins); !xs.empty(); xs++, xd++) {
					_pwr_chg[_ch] -= norm(*xd);
					run(xd, xs);
					_pwr_chg[_ch] += norm(*xd);
				}
			}
		}
	}
	void shiftUpA(long b0, long b1)
	{
		float hz = (float)b1*_bin_to_hz;
		float part_note = 48.0f*NoteFreq::inv_log_c0_hz*(logf(hz)-NoteFreq::log_c0_hz);
		float cf = NoteFreq::c0_hz*powf(2.0f, floorf(part_note)*(1.0f/48.0f));
		float cb = cf*_hz_to_bin;

		long v1 = b1;
		while(1) {
			long bin_shift = (long)(cb*_shift_mult+0.5f);
			float next_cb = cb*(float)0.98566319864018758778;
			long v0 = (long)((next_cb+cb)*0.5f);
			if(v0 <= v1) {
				_phase() = _phase(bin_shift) * AmpProcess::_amp;
				run(v0+bin_shift, v0, v1-v0+1);
				v1 = v0-1;
				if(v1 < b0) break;
			}
			cb = next_cb;
		}
	}

	void shiftDownA(long b0, long b1)
	{
		float hz = (float)b0*_bin_to_hz;
		float part_note = 48.0f*NoteFreq::inv_log_c0_hz*(logf(hz)-NoteFreq::log_c0_hz);
		float cf = NoteFreq::c0_hz*powf(2.0f, ceilf(part_note)*(1.0f/48.0f));
		float cb = cf*_hz_to_bin;

		long v0 = b0;
		while(1) {
			long bin_shift = (long)(cb*_shift_mult+0.5f);
			float next_cb = cb*(float)1.0145453349375237462;
			long v1 = (long)((next_cb+cb)*0.5f);
			if(v0 <= v1) {
				_phase() = _phase(bin_shift) * AmpProcess::_amp;
				run(v0+bin_shift, v0, v1-v0+1);
				v0 = v1+1;
				if(v0 > b1) break;
			}
			cb = next_cb;
		}
	}

	void shiftUpB(long b0, long b1)
	{
		// process in reverse
		// try to align shifts with peaks in the data
		long v0[3], v1[3];
		PeakFind p[3];
		v1[0] = b1;
		while(1) {
			v0[0] = max(v1[0]-_type_b_width, b0);
			v1[1] = v0[0]-1;
			v0[1] = max(v1[1]-_type_b_width, b0);
			v1[2] = v0[1]-1;
			v0[2] = max(v1[2]-_type_b_width, b0);

			for(int i = 0; i < 3; i++) p[i](_b, v0[i], v1[i]);

			// decide whether to use p[0] or p[1] as the peak
			long cent;
			if(p[1].max_pwr > p[0].max_pwr && p[1].max_pwr > p[2].max_pwr)
				cent = p[1].max_bin;
			else cent = p[0].max_bin;

			long bin_shift = (long)((float)cent*_shift_mult+0.5f);
			v0[0] = max(cent-_type_b_width, b0);
			_phase() = _phase(bin_shift) * AmpProcess::_amp;
			run(v0[0]+bin_shift, v0[0], v1[0]-v0[0]+1);

			if(v0[0] <= b0) break;
			v1[0] = v0[0]-1;
		}
	}

	void shiftDownB(long b0, long b1)
	{
		// try to align shifts with peaks in the data
		long v0[3], v1[3];
		PeakFind p[3];
		v0[0] = b0;
		while(1) {
			v1[0] = min(v0[0]+_type_b_width, b1);
			v0[1] = v1[0]+1;
			v1[1] = min(v0[1]+_type_b_width, b1);
			v0[2] = v1[1]+1;
			v1[2] = min(v0[2]+_type_b_width, b1);

			for(int i = 0; i < 3; i++) p[i](_b, v0[i], v1[i]);

			// decide whether to use p[0] or p[1] as the peak
			long cent;
			if(p[1].max_pwr > p[0].max_pwr && p[1].max_pwr > p[2].max_pwr)
				cent = p[1].max_bin;
			else cent = p[0].max_bin;

			long bin_shift = (long)((float)cent*_shift_mult+0.5f);
			v1[0] = min(cent+_type_b_width, b1);
			_phase() = _phase(bin_shift) * AmpProcess::_amp;
			run(v0[0]+bin_shift, v0[0], v1[0]-v0[0]+1);

			if(v1[0] >= b1) break;
			v0[0] = v1[0]+1;
		}
	}

	void run(long b0, long b1)
	{

		if(_do_const_shift) {
			// constant shift is easy
			long n_bins = b1-b0+1;
			long src_bin, dst_bin;
			
			if(_dest_mode) {
				dst_bin = b0;
				src_bin = b0-_const_bin_shift;
			}
			else {
				src_bin = b0;
				dst_bin = b0+_const_bin_shift;
			}
			run(dst_bin, src_bin, n_bins);
		}
		else {
			// scaled shift is messy

			if(_dest_mode) {
				b1 = (long)((float)b1*_inv_freq_scale+0.5f);
				b0 = (long)((float)b0*_inv_freq_scale+0.5f);
			}

			if(b1 < 1) return;
			if(b0 < 1) b0 = 1;

			if(_reverse) {
				// shift up
				if(b1 >= _type_a_low_bin)
					shiftUpA(max(_type_a_low_bin, b0), b1);
				if(b0 < _type_a_low_bin)
					shiftUpB(b0, min(_type_a_low_bin-1, b1));
			}
			else {
				// shift down
				if(b0 < _type_a_low_bin)
					shiftDownB(b0, min(_type_a_low_bin-1, b1));
				if(b1 >= _type_a_low_bin)
					shiftDownA(max(_type_a_low_bin, b0), b1);
			}
		}
	}
	float pwrChg(int ch) { return _pwr_chg[ch] + _src_pwr[ch]*(_mix_amp*_mix_amp-1.0f); }
};


//-------------------------------------------------------------------------------------------------
int ShiftDispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val)
{
	SplitParam<2> split(val);
	if(split.i_part) {
		// constant freq shift
		float hz = GetFreqShift(split.f_part);
		return sprintf(
			text,
			"%.0f Hz", 
			b? b->guessRoundHz(hz):hz
		);
	}
	else {
		// scaled shift
		return sprintf(text, "%.2f notes",  GetOctShift(split.f_part)*12.0f);
	}
}


//- SHIFT ADD -------------------------------------------------------------------------------------
class ShiftAddFx;
template <> inline void ShiftProcess<ShiftAddFx>::run(pcplxf dst, pcplxf src)
{ 
	cplxf s = *src;
	_src_pwr[_ch] += norm(s);
	*src = s * _mix_amp;
	*dst = *dst + s * _phase();		
}

//-------------------------------------------------------------------------------------------------
class ShiftAddFx : public FFTFx
{
public:
	bool _dest_mode;

	ShiftAddFx(const char* name, bool dest_mode) : FFTFx(name)
		{ _dest_mode = dest_mode; }

	virtual void process(DtBlkFx* b) { run(b, ShiftProcess<ShiftAddFx>(b, _dest_mode)); }

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return ShiftDispVal(b, text, val); }

	virtual bool ampMixMode() { return true; }
};

ShiftAddFx g_shift_add_fx("ShiftAdd", /*dst mode*/false);
ShiftAddFx g_shift_addd_fx("ShiftAddD", /*dst mode*/true);


//- SHIFT REPL ------------------------------------------------------------------------------------
class ShiftReplFx;
template <> inline void ShiftProcess<ShiftReplFx>::run(pcplxf dst, pcplxf src)
{
	*dst = *dst * _mix_amp + *src * _phase();
}

//-------------------------------------------------------------------------------------------------
class ShiftReplFx : public FFTFx
{
public:
	bool _dest_mode;

	ShiftReplFx(const char* name, bool dest_mode) : FFTFx(name)
		{ _dest_mode = dest_mode; }

	virtual void process(DtBlkFx* b) { run(b, ShiftProcess<ShiftReplFx>(b, _dest_mode)); }

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return ShiftDispVal(b, text, val); }
};

ShiftReplFx g_shift_repl_fx("ShiftRepl", /*dst mode*/false);
ShiftReplFx g_shift_repld_fx("ShiftReplD", /*dst mode*/true);

//*************************************************************************************************
class AutoHarmFx : public FFTFx
{
public:
	AutoHarmFx() : FFTFx("AutoHarm") {}

	virtual void process(DtBlkFx* b)
	{
		FxParams& params = b->currFxParams();
		long stop_bin = b->_freq_fft_n/2-1;
		if(params.bin[0] > params.bin[1]) stop_bin = params.bin[0];

		AmpProcess amp(b);
		AutoHarmMaskProcess<AmpProcess> harm(amp, b, params, params.val, /*multiplier*/1.0f);
		run(b, harm, /*split mask*/false, /*start_bin*/1, stop_bin);
	}

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return HarmDispVal(text, val); }

} g_auto_harm_fx;


//*************************************************************************************************
class MirrorProcess : public AmpProcess
{
public:
	float _s[2];

	MirrorProcess(DtBlkFx* b) : AmpProcess(b)
	{
		float v = b->currFxParams().val;
		_s[0] = _amp * (1-v);
		_s[1] = _amp * v;
	}

	void run(long b0, long b1)
	{
		for(int ch = 0 ; ch < AUDIO_CHANNELS; ch++) {
			pcplxf xa = _b->FFTdata(ch)+b0;
			pcplxf xb = _b->FFTdata(ch)+b1;
			while(xa.r <= xb.r) {
				_pwr_chg[ch] -= norm(*xa);
				_pwr_chg[ch] -= norm(*xb);
				cplxf ta = *xa;
				cplxf tb = *xb;
				*xa = ta*_s[0] + tb*_s[1];
				*xb = tb*_s[0] + ta*_s[1];
				_pwr_chg[ch] += norm(*xa);
				_pwr_chg[ch] += norm(*xb);
				xa++;
				xb--;
			}
		}
	}
};


//-------------------------------------------------------------------------------------------------
class MirrorFx : public FFTFx
{
public:
	MirrorFx() : FFTFx("Mirror") { }
	virtual void process(DtBlkFx* b) { run(b, MirrorProcess(b)); }

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return sprintf(text, "%.0f%%", val*100.0f); }

} g_mirror_effect;


//*************************************************************************************************
//-------------------------------------------------------------------------------------------------
class HarmMatchProcess : public AmpProcess
{
public:
	// 2 harmonic sequences that we linearly interpolate against
	float* _harms[2];

	// number of harmonics per sequence
	int _n_harms;

	// scale factors (0..1) to apply to harm sequences
	float _harms_interp[2];

	// power scaling to apply to harmonics
	float _pwr_scale[AUDIO_CHANNELS];

	// true=copy fundamental harmonic, false=normal power matching
	bool _copy_mode;

	// amp to apply to original data
	float _orig_amp;

	// our parent must be HarmMask
	HarmMaskProcess<HarmMatchProcess> *_parent;


	void prepare(HarmMaskProcess<HarmMatchProcess>* parent)
	{
		AmpProcess::prepare(parent);

		_parent = parent;

		long max_bin = _b->_freq_fft_n/2-1;

		// determine bin range of fundamental
		long cf = (long)(parent->_freq+0.5f);
		long w2 = parent->_half_width;
		long f0 = limit_range(cf - w2, 1, max_bin);
		long f1 = min(cf + w2, max_bin);

		// make sure range is valid
		if(f0 > f1) return;

		// get pwr scaling to use
		float harm0_pwr = harmPwr(0);
		if(harm0_pwr <= 0) return;

		for(int ch = 0; ch < AUDIO_CHANNELS; ch++) {
			_pwr_scale[ch] = GetPwr(_b->FFTdata(ch), f0, f1)/harm0_pwr;
			if(_copy_mode) _pwr_scale[ch] *= _amp*_amp;
		}
	}

	// return interpolated harmonic power from table
	float harmPwr(int harmonic)
		{ return _harms[0][harmonic]*_harms_interp[0]+_harms[1][harmonic]*_harms_interp[1]; }


	HarmMatchProcess(
		DtBlkFx* b,
		HarmData harm_data
	) : AmpProcess(b)
	{
		// default for pwr scale
		for(int ch = 0; ch < AUDIO_CHANNELS; ch++) _pwr_scale[ch] = 1.0f;
		
		_n_harms = harm_data.n_harms;

		// grab the fx params
		FxParams& params = b->currFxParams();
		SplitParam<2> val(params.val);

		//
		_copy_mode = val.i_part?true:false;

		// find which harmonic sequences to interpolate from
		int max_table = harm_data.n_tables-1;
		float ftable_i;
		_harms_interp[1] = modff(val.f_part*(float)max_table, &ftable_i);
		_harms_interp[0] = 1-_harms_interp[1];

		int table_i = limit_range((int)ftable_i, 0, max_table);
		_harms[0] = harm_data.data+table_i*harm_data.n_harms;

		if(table_i < max_table) table_i++;
		_harms[1] = harm_data.data+table_i*harm_data.n_harms;

		// determine whether we need to mix or replace
		_orig_amp = max(1.0f-_amp, 0.0f);
	}

	void run(long b0, long b1)
	{
		long curr_harm = (long)_parent->_curr_harm;
		if(curr_harm < 0 || curr_harm >= _n_harms) return;

		// get required power for this harmonic
		float harm_pwr = harmPwr(curr_harm);

		if(_copy_mode) {
			// find copy position for fundamental
			long cent_offs = (long)(_parent->_freq - _parent->_curr_cent + 0.5f);
			long f0 = cent_offs + b0;
			long f1 = cent_offs + b1;

			// make sure bin range is within limits
			if(f1 >= _parent->_max_bin) {
				b1 += _parent->_max_bin - f1;
				f1 = _parent->_max_bin;
			}
			if(f0 < 1) {
				b0 += 1-f0;
				f0 = 1;
			}

			// do nothing if inverted range
			if(f0 > f1 || b0 > b1) return;

			// do the copy for each channel from the fundamental
			for(int ch = 0; ch < AUDIO_CHANNELS; ch++) {
				// copy mode
				pcplxf src = _b->FFTdata(ch)+f0;

				// scaling for harmonic
				float scale = sqrtf(harm_pwr*_pwr_scale[ch]);

				// power change due to this effect
				float pwr_chg = 0;

				if(_orig_amp) {
					// mix mode, mix source & destination
					// use scaling on source as per "scale mode"

					for(pcplxf_range x(_b->FFTdata(ch), b0, b1+1); !x.empty(); x++) {
						pwr_chg -= norm(*x);
						*x = (*x) * _orig_amp + (*src) * scale;
						pwr_chg += norm(*x);
						src++;
					}
				}
				else {
					// not mix mode, replace destination
					for(pcplxf_range x(_b->FFTdata(ch), b0, b1+1); !x.empty(); x++) {
						pwr_chg -= norm(*x);
						*x = (*src) * scale;
						pwr_chg += norm(*x);
						src++;
					}
				}
				_pwr_chg[ch] += pwr_chg;
			}
		}
		else {

			// scale mode
			for(int ch = 0; ch < AUDIO_CHANNELS; ch++) {
				// find what we need to scale the existing data by to match the harmonic power
				float orig_pwr = GetPwr(_b->FFTdata(ch), b0, b1);
				float target_pwr = _pwr_scale[ch]*harm_pwr;

				// attempt to match to harmonic pwr
				float scale = _orig_amp + MatchPwr(_amp, target_pwr, orig_pwr);

				for(pcplxf_range x(_b->FFTdata(ch), b0, b1+1); !x.empty(); x++)
					*x = (*x) * scale;

				// store power change
				_pwr_chg[ch] += (scale*scale-1)*orig_pwr;
			}
		}
	}
};

//-------------------------------------------------------------------------------------------------
template <class T> void runHarmMask(T& t, DtBlkFx* b)
//
// run effects that require a harm mask
// create a default harm mask if previous effect isn't harmmask
//
{
	// attempt to use previous effect as harmonic mask
	if(runHarmMask(b, t, /*split mask*/true, 
			/*start bin*/1, /*stop bin*/b->_freq_fft_n/2-1)) return;

	// if the previous effect wasn't a harmonic mask, create one using a constant fx val param
	AutoHarmMaskProcess<T>
		def_mask(t, b, b->currFxParams(), /*fx_val*/0.2499f, /*freq mult*/1.0f);
	
	// run without slit mask since we are using freq's for doing the auto-harm
	run(b, def_mask, /*split mask*/false);
}

//-------------------------------------------------------------------------------------------------
class HarmMatchFx : public FFTFx
{
public:
	HarmData _data;
	HarmMatchFx(const char* name, HarmData data) : FFTFx(name) { _data = data; }
	virtual void process(DtBlkFx* b)
	{
		// if the previous effect is a harmonic mask then use it (so that the user can control
		// harmonic width) otherwise default to processing all harmonics with a width of 50%
		HarmMatchProcess match(b, _data);
		runHarmMask(match, b);
	}

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
	{
		SplitParam<2> v(val);
		return sprintf(text, "%s/%.0f", v.i_part?"copy":"cabbage", v.f_part*100);
	}
	virtual bool ampMixMode() { return true; }
};

extern HarmData sweep1_coeff;
HarmMatchFx g_sweep1_fx("Triangles", sweep1_coeff);

extern HarmData sweep2_coeff;
HarmMatchFx g_sweep2_fx("Squares", sweep2_coeff);

extern HarmData sweep3_coeff;
HarmMatchFx g_sweep3_fx("Saws", sweep3_coeff);

extern HarmData sweep4_coeff;
HarmMatchFx g_sweep4_fx("Pointy", sweep4_coeff);

extern HarmData sweep5_coeff;
HarmMatchFx g_sweep5_fx("Sweep", sweep5_coeff);

//*************************************************************************************************

//-------------------------------------------------------------------------------------------------
class HarmShiftProcess : public AmpProcess
{
public:
	bool _repitch_mode;

	// frequency scaling to apply to bins
	float _freq_scale;

	// multiplier to get bin shift
	float _shift_mult;

	// our parent must be HarmMask
	HarmMaskProcess<HarmShiftProcess> *_parent;

	// scaling to apply to src data
	float _src_amp;

	// _src_amp**2 - 1
	float _src_pwr_scale;

	// fft len / 2
	long _max_bin;

	int _ch;

	// phase correction for shifted bins
	PhaseCorrect _phase;

	void prepare(HarmMaskProcess<HarmShiftProcess>* parent)
	{
		AmpProcess::prepare(parent);
		_parent = parent;

		if(_repitch_mode) {
			// scale so that fundamental is moved to the destination note
			float note_offs_c0 = _b->currFxParams().val*BlkFxParam::noteSpan();
			float denom = _parent->_freq * _b->getSampleRate();
			if(denom > 0) 
				_freq_scale = NoteOffsToHz(note_offs_c0)*(float)_b->_freq_fft_n / denom;
			else _freq_scale = 1.0f;
		}
		else {
			// get multiplier from octave shift
			_freq_scale = powf(2.0f, GetOctShift(_b->currFxParams().val));
		}
		_shift_mult = _freq_scale - 1.0f;
	}

	HarmShiftProcess(
		DtBlkFx* b,
		bool repitch_mode
	) : AmpProcess(b)
	{
		_repitch_mode = repitch_mode;
		_phase.init(b);

		// determine whether we need to mix or replace
		if(_amp < 1.0f) _src_amp = 1.0f - _amp;
		else _src_amp = 0.0f;
		
		_src_pwr_scale = _src_amp*_src_amp-1.0f;

		_max_bin = b->_freq_fft_n/2;
	}

	bool reverse() { return _freq_scale > 1.0f; }

	void run(pcplxf_range dst, pcplxf_range src, int step /*-1 or 1*/)
	{
		float src_pwr = 0.0f;

		// process any src data that has out of range dst
		while(!src.empty()) {
			src_pwr += norm(*src);
			*src = (*src) * _src_amp;
			src += step;
		}
		// process src to dst
		while(!dst.empty()) {
			cplxf s = *src;
			src_pwr += norm(s);
			*src = s * _src_amp;

			_pwr_chg[_ch] -= norm(*dst);
			*dst = (*dst) + s * _phase();
			_pwr_chg[_ch] += norm(*dst);

			src += step;
			dst += step;
		}

		_pwr_chg[_ch] += src_pwr*_src_pwr_scale;
	}

	void run(long b0, long b1)
	{
		// determine offset to apply to bins
		long bin_shift = (long)(_parent->_curr_cent * _shift_mult + 0.5f);

		_phase() = _phase(bin_shift) * _amp;

		// destination bins
		long d0 = bin_shift + b0;
		long d1 = bin_shift + b1;

		// s0..s1 is source bin range corresponding to out-of-range destination
		long s0, s1;
		if(reverse()) {
			s0 = b1;
			s1 = b1;
			if(d1 > _max_bin) {
				s0 = max(b1+_max_bin-d1, b0-1);
				d1 = _max_bin;
				if(d0 > d1) d0 = d1+1;
			}
		}
		else {
			s0 = b0;
			s1 = b0;
			if(d0 < 1) {
				s1 = min(b0+1-d0, b1+1);
				d0 = 1;
				if(d0 > d1) d1 = d0-1;
			}
		}

		// process each channel
		for(_ch = 0; _ch < AUDIO_CHANNELS; _ch++) {
			pcplxf x1 = _b->FFTdata(_ch);

			if(reverse()) {
				// shifting up in frequency
				run(/*dst*/pcplxf_range(x1, d1, d0-1), /*out of rng src*/pcplxf_range(x1, s1, s0), /*step*/-1);
			}
			else {
				// shifting down in frequency
				run(/*dst*/pcplxf_range(x1, d0, d1+1), /*out of rng src*/pcplxf_range(x1, s0, s1), /*step*/ 1);
			}
		}
	}
};

//-------------------------------------------------------------------------------------------------
class HarmShiftFx : public FFTFx
{
public:
	HarmShiftFx() : FFTFx("HarmShift") {}

	virtual void process(DtBlkFx* b)
	{
		HarmShiftProcess shift(b, /*repitch*/false);
		runHarmMask(shift, b);
	}

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
	{
		// display octave shift as notes (semitones) shift
		float notes = GetOctShift(val)*12.0f;
		return sprintf(text, "%.1f notes", notes);
	}

	virtual bool ampMixMode() { return true; }
} g_harm_shift;


//-------------------------------------------------------------------------------------------------
class HarmRepitchFx : public FFTFx
{
public:
	HarmRepitchFx() : FFTFx("HarmRepitch") {}

	virtual void process(DtBlkFx* b)
	{
		HarmShiftProcess shift(b, /*repitch*/true);
		runHarmMask(shift, b);
	}

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
	{
		float offs_c0 = val*BlkFxParam::noteSpan();
		return NoteToTxt(text, offs_c0);
	}

	virtual bool ampMixMode() { return true; }
} g_harm_repitch;

//*************************************************************************************************
class ResampleProcess : public AmpProcess
{
public:
	PhaseCorrect _phase;

	// max bin that can be processed without overflowing
	long _max_bin;

	// frequency multiply
	float _freq_mult;

	// shifting multiply & step
	float _fshift_mult, _fshift_step;

	// true if we are shifting up in frequency
	bool _shift_up;

	// process in reverse
	bool reverse() { return _shift_up; }

	// amplification to apply to source
	float _src_amp;

	float _src_pwr_scale;

	ResampleProcess(DtBlkFx* b) : AmpProcess(b)
	{
		_phase.init(b);

		// mult freqs by this much
		_freq_mult = powf(2.0f, GetOctShift(b->currFxParams().val));

		// maximum bin that can be processed without exceeding fft length
		_max_bin = (long)((float)(b->_freq_fft_n/2-1)/_freq_mult+0.5f);

		// multiply a frequency to get shift amount
		_fshift_mult = _freq_mult-1.0f;

		// whether we're shifting up or down
		_shift_up = (_freq_mult > 1.0f);

		// shift step
		if(_shift_up) _fshift_step = -_fshift_mult;
		else _fshift_step = _fshift_mult;

		if(_amp < 1.0f) _src_amp = 1.0f-_amp;
		else _src_amp = 0.0f;

		// power scaling after applying source scaling
		_src_pwr_scale = _src_amp*_src_amp-1.0f;
	}

	void run(
		long b_start,		// start bin
		long b_end,			// end bin
		long n_src_only,	// number of bins that have out of range destination
		int step			// 1=fwd, -1=rev
	)
	{
		int ch;
		pcplxf src[AUDIO_CHANNELS];
		float src_pwr[AUDIO_CHANNELS];
		for(ch = 0; ch <AUDIO_CHANNELS; ch++) {
			src[ch] = _b->FFTdata(ch) + b_start;
			src_pwr[ch] = 0;
		}

		// process any src data that has out of range dst
		float* src0_only_end = src[0].r - n_src_only;
		while(src[0].r != src0_only_end) {
			for(ch = 0; ch < AUDIO_CHANNELS; ch++) {
				src_pwr[ch] += norm(*src[ch]);
				*src[ch] = (*src[ch]) * _src_amp;
				src[ch] += step;
			}
		}

		// process src to dst
		float fshift = _fshift_mult*(float)(b_start - n_src_only);
		float* src0_end = _b->FFTdata(0).r + b_end + step;
		while(src[0].r != src0_end) {
			long shift = (long)(fshift+0.5f);
			_phase() = _phase(shift) * _amp;

			for(ch = 0; ch < AUDIO_CHANNELS; ch++) {
				pcplxf dst = src[ch]+shift;
				cplxf s = *src[ch];

				src_pwr[ch] += norm(s);
				*src[ch] = s * _src_amp;

				_pwr_chg[ch] -= norm(*dst);
				*dst = (*dst) + s * _phase();
				_pwr_chg[ch] += norm(*dst);
				src[ch] += step;
			}
			fshift += _fshift_step;
		}

		for(ch = 0; ch < AUDIO_CHANNELS; ch++)
			_pwr_chg[ch] += src_pwr[ch]*_src_pwr_scale;
	}

	void run(long b0, long b1)
	{

		// variables from AmpProcess: _b
		if(_shift_up) {
			// shifting up in frq, process backwards
						
			// find number of bins that are shifted out of range
			int n_src_only = 0;
			if(b1 > _max_bin)
				n_src_only = min(b1-_max_bin, b1-b0+1);

			run(b1, b0,			// reverse
				n_src_only,		// number of out of rng bins
				-1				// source step
			);

		}
		else {
			// don't worry about bins ending up at 0 hz
			// shifting down in frq, process forward
			run(b0, b1,			// forwards
				0,				// all bins have destination
				1				// source step
			);
		}
	}
};

//-------------------------------------------------------------------------------------------------
class ResampleFx : public FFTFx
{
public:
	ResampleFx() : FFTFx("Resample") {}

	virtual void process(DtBlkFx* b) { run(b, ResampleProcess(b)); }

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
	{
		// display octave shift as notes (semitones) shift
		float notes = GetOctShift(val)*12.0f;
		return sprintf(text, "%.1f notes", notes);
	}

	virtual bool ampMixMode() { return true; }
} g_resample_fx;


//*************************************************************************************************
#ifdef STEREO
class VocodeFx : public FFTFx
// vocoder is a stereo effect
// it uses the frequency envelope from channel 0 (left) to modulate channel 1 (right)
{
public:
	VocodeFx() : FFTFx("Vocode") {}

	static float getSegs(float params_val) { return min(params_val*(399.0f/.875f)+1.0f, 400.0f); }

	// top 12.5% of value mixes in spectrum multiply
	static float getMultFrac(float params_val) { return max((params_val-0.875f) * 8.0f, 0.0f); }

	virtual void process(DtBlkFx* b)
	{
		// get param set
		FxParams& params = b->currFxParams();

		float amp = params.amp;
		float mult_frac = getMultFrac(params.val);

		// fraction of voc/mult in output
		float wet_frac = min(amp, 1.0f);

		float voc_amp = amp*(1-mult_frac);
		float voc_mixback = 1-wet_frac*(1-mult_frac);

		float mult_amp = amp*mult_frac;
		float mult_mixback = 1-wet_frac*mult_frac;

		// fraction of power in ch1 that will be adjusted
		float voc_ch1_in_pwr_frac = 0.0f;
		float mult_ch1_in_pwr_frac = 0.0f;

		// ch0 input power processed
		float voc_ch0_in_pwr = 0.0f;
		float mult_ch0_in_pwr = 0.0f;

		// do vocode (envelope match)
		if(voc_amp > 0.0f) {

			// get src & dst bin ranges
			Array<long, 2> src_b;
			Array<long, 2> dst_b;
			if(params.bin[0] > params.bin[1]) {
				// freqA > freqB, use range to select source
				src_b[0] = params.bin[1];
				src_b[1] = params.bin[0];
				dst_b[0] = 1;
				dst_b[1] = b->_freq_fft_n/2;
			}
			else {
				// freqB > freqA, use range to select dest
				dst_b = params.bin;
				src_b[0] = 1;
				src_b[1] = b->_freq_fft_n/2;
			}

			// get number of bins in src & dst
			long dst_n = dst_b[1]-dst_b[0];
			if(!dst_n) return;
			long src_n = src_b[1]-src_b[0];
			if(!src_n) return;

			// get src & dst data ranges (the range corresponds to each block)
			pcplxf_range src;
			src.start() = b->FFTdata(0)+src_b[0];

			pcplxf_range dst;
			dst.start() = b->FFTdata(1) + dst_b[0];
			float* dst_end = b->FFTdata(1).r + dst_b[1];

			// get number of segments (arbitrary scaling)
			float n_segs = limit_range(getSegs(params.val), 1.0f, (float)dst_n);

			float fsrc_stp = (float)src_n/n_segs;
			float fdst_stp = (float)dst_n/n_segs;

			float fsrc_b = (float)src_b[0];
			float fdst_b = (float)dst_b[0];

			float total_ch1_in_pwr = 0.0f;
			float total_ch1_out_pwr = 0.0f;

			// run in a loop over the number of segments to match pwr in each dst segment with src
			while(1) {
				// find the src & dst ranges
				fdst_b += fdst_stp;
				fsrc_b += fsrc_stp;

				pcplxf src_next = (long)fsrc_b+b->FFTdata(0);

				src.e = src_next.r;
				dst.e = (long)fdst_b+b->FFTdata(1).r;
				if(dst.e > dst_end) break;

				// get the power from the src segment (or from src start bin if empty)
				float src_pwr;
				if(src.empty()) src_pwr = norm(*src);
				else {
					src_pwr = GetPwr(src);
					voc_ch0_in_pwr += src_pwr;
				}

				float dst_in_pwr = GetPwr(dst);
				float dst_scale = MatchPwr(voc_amp, src_pwr, dst_in_pwr)+voc_mixback;

				// scale the dst segment to match the src segment
				for(;!dst.empty(); dst++) *dst = (*dst) * dst_scale;

				// next segment
				src.start() = src_next;

				total_ch1_in_pwr += dst_in_pwr;
				total_ch1_out_pwr += dst_in_pwr*dst_scale*dst_scale;
			}

			if(b->_chan[1].total_out_pwr > 1e-20f)
				voc_ch1_in_pwr_frac = total_ch1_in_pwr/b->_chan[1].total_out_pwr;

			// adjust output pwr for what we've just done
			b->_chan[1].total_out_pwr += total_ch1_out_pwr-total_ch1_in_pwr;
		}

		// multiply mode
		// TODO: make this efficient
		// note, the freq range multiplied will be different to the vocoded range if freqb < freqa

		if(mult_amp > 0.0f) {
			// lerp mix between dst & dst*src based on mixback

			Array<long, 2> bin = params.bin;
			if(bin[0] > bin[1]) swap(bin[0], bin[1]);
			pcplxf_range ch0_(b->FFTdata(0), bin[0], bin[1]+1), ch0;
			pcplxf ch1_ = b->FFTdata(1)+bin[0], ch1;

			// pass1: ch0 = ch0*ch1
			float ch0_out_pwr = 0.0f;
			for(ch0 = ch0_, ch1 = ch1_; !ch0.empty(); ch0++, ch1++) {
				mult_ch0_in_pwr += norm(*ch0);
				(*ch0) = (*ch0)*(*ch1);
				ch0_out_pwr += norm(*ch0);
			}

			// pass2: power match to ch0 in pwr & mix to output
			float ch0_scale = MatchPwr(mult_amp, mult_ch0_in_pwr, ch0_out_pwr);
			float ch1_out_pwr = 0.0f;
			float ch1_in_pwr = 0.0f;
			for(ch0 = ch0_, ch1 = ch1_; !ch0.empty(); ch0++, ch1++) {
				ch1_in_pwr += norm(*ch1);
				(*ch1) = ch0_scale*(*ch0) + mult_mixback*(*ch1);
				ch1_out_pwr += norm(*ch1);
			}
			if(b->_chan[1].total_out_pwr > 1e-20f)
				mult_ch1_in_pwr_frac = ch1_in_pwr/b->_chan[1].total_out_pwr;

			b->_chan[1].total_out_pwr += ch1_out_pwr-ch1_in_pwr;
		}

		// replace part (or all) of ch 1 input power with ch 0 input power, lerp since spectrum ranges may be
		// different between voc & mult

		// find out fraction of input spectrum we're matching to
		float match_frac = lin_interp(mult_frac, voc_ch1_in_pwr_frac, mult_ch1_in_pwr_frac);

		// remove some fraction of original ch1 pwr & replace with ch0 pwr
		b->_chan[1].total_in_pwr *= 1-wet_frac*match_frac;
		b->_chan[1].total_in_pwr += wet_frac*lin_interp(mult_frac, voc_ch0_in_pwr, mult_ch0_in_pwr);

		// copy the dst back to channel 0 (src)
		memcpy(b->FFTdata(0).r, b->FFTdata(1).r, b->_freq_fft_n*sizeof(float));
		b->_chan[0].total_in_pwr = b->_chan[1].total_in_pwr;
		b->_chan[0].total_out_pwr = b->_chan[1].total_out_pwr;
	}

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
	{

		if(getMultFrac(val) > 0.0f)
			return sprintf(text, "%.0f%% mult", getMultFrac(val)*100.0f);
		return sprintf(text, "%.0f segs", getSegs(val));
	}

	virtual bool ampMixMode() { return true; }
} g_vocode;

//*************************************************************************************************
class HarmMatch2Process : public AmpProcess
{
public:
	// our parent must be HarmMaskProcess
	HarmMaskProcess<HarmMatch2Process> *_parent;

	// mixback amp
	float _mix_back;

	// max FFT bin
	long _max_bin;

	// how much chg in dst power
	float _dst_pwr_chg;

	int _src_ch, _dst_ch;

	// expect to be filled in by caller
	HarmParam _src_harm;

	void prepare(HarmMaskProcess<HarmMatch2Process>* parent)
	{
		AmpProcess::prepare(parent);
		_parent = parent;
	}

	HarmMatch2Process(DtBlkFx* b, int src_ch, int dst_ch) : AmpProcess(b)
	{
		_src_ch = src_ch;
		_dst_ch = dst_ch;

		_dst_pwr_chg = 0.0f;

		// max FFT bin
		_max_bin = _b->_freq_fft_n/2-1;

		// determine whether we need to mix or replace
		_mix_back = max(1.0f-_amp, 0.0f);
	}

	void run(long b0, long b1)
	// scale the power of the destination harmonic to match the src harmonic
	{
		//
		long curr_harm = (long)_parent->_curr_harm;
		if(curr_harm < 0) return;

		// power in destination harmonic
		float dst_in_pwr = GetPwr(_b->FFTdata(_dst_ch), b0, b1);

		// determine src harmonic
		long s0, s1;
		s0 = (long)(_src_harm._offs + curr_harm*_src_harm._spacing) - _src_harm._half_width;
		s1 = s0+_src_harm._width;
		if(s0 < 1) s0 = 1;
		if(s1 > _max_bin) s1 = _max_bin;

		// get src harmonic pwr
		float src_pwr = dst_in_pwr;
		if(s1 >= s0) src_pwr = GetPwr(_b->FFTdata(_src_ch), s0, s1);

		// attempt to match dst pwr to src pwr
		float scale = MatchPwr(_amp, src_pwr, dst_in_pwr)+_mix_back;

		// scale the destination data
		for(pcplxf_range x(_b->FFTdata(_dst_ch), b0, b1+1); !x.empty(); x++) *x = (*x) * scale;

		_dst_pwr_chg += dst_in_pwr*(scale*scale-1.0f);
	}

	// power is corrected in Fx class
	float pwrChg(int ch) { return 0.0f; }
};

//-------------------------------------------------------------------------------------------------
class HarmMatch2Fx : public FFTFx
{
public:
	bool _RL_mode;
	HarmMatch2Fx(const char* name, bool RL_mode) : FFTFx(name)
	{
		_RL_mode = RL_mode;
	}

	virtual void process(DtBlkFx* b)
	{
		// RL_mode=0: src ch is left, dst ch is right
		// RL_mode=1: src ch is right, dst ch is left

		int src_ch = 0, dst_ch = 1;
		if(_RL_mode) swap(src_ch, dst_ch);

		HarmMatch2Process match(b, src_ch, dst_ch);

		FxParams& params = b->currFxParams();

		// find peaks in src & dst data
		Array<long, 2> dst_bin = params.bin, src_bin;
		if(dst_bin[0] > dst_bin[1]) swap(dst_bin[0], dst_bin[1]);

		float fx_val = params.val;
		src_bin = dst_bin;

		// chk whether the previous effect is a harmonic mask, if so use its params to control the source
		FxParams* prev_params = b->prevFxParams();
		if(prev_params) {
			if(getHarmMaskFreqMult(prev_params->fft_fx, /*out*/ match._src_harm._multiply)) {
				// use previous mask as src bin control
				src_bin = prev_params->bin;
				if(src_bin[0] > src_bin[1]) swap(src_bin[0], src_bin[1]);
				fx_val = prev_params->val;
			}
		}
		float src_peak = (float)PeakFind1Ch(b->FFTdata(src_ch), src_bin[0], src_bin[1]);
		match._src_harm.init(fx_val, src_peak);

		long dst_peak = PeakFind1Ch(b->FFTdata(dst_ch), dst_bin[0], dst_bin[1]);

		run_(
			b,
			HarmMaskProcess<HarmMatch2Process>(match, params.val, (float)dst_peak),
			/*start bin*/1, /*final bin*/b->_freq_fft_n/2-1
		);

		// always match overall pwr to left ch (no matter which is src & dst)
		b->_chan[1].total_in_pwr = b->_chan[0].total_in_pwr;

		// adjust dest ch pwr out pwr
		b->_chan[dst_ch].total_out_pwr += match._dst_pwr_chg;

		// copy dst data back to src (both channels are now "dst")
		memcpy(b->FFTdata(src_ch).r, b->FFTdata(dst_ch).r, b->_freq_fft_n*sizeof(float));
		b->_chan[src_ch].total_in_pwr = b->_chan[dst_ch].total_in_pwr;
		b->_chan[src_ch].total_out_pwr = b->_chan[dst_ch].total_out_pwr;
	}

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return HarmDispVal(text, val); }

	virtual bool ampMixMode() { return true; }
}
	g_harmmatchLR_fx("HarmMatchLR", /*R-L mode*/false),
	g_harmmatchRL_fx("HarmMatchRL", /*R-L mode*/true);


#endif

//*************************************************************************************************
FFTFx *g_fft_fx_table[] =
{
	&g_filter_fx,
	&g_contrast_fx,
	&g_smear_fx,
	&g_thresh_fx,
	&g_clip_fx,
	&g_mirror_effect,
	&g_resample_fx,
	&g_shift_add_fx,
	&g_shift_addd_fx,
	&g_shift_repl_fx,
	&g_shift_repld_fx,
	&g_harm_shift,
	&g_harm_repitch,
	&g_harm_filt_fx,
	&g_auto_harm_fx, 
	&g_sweep1_fx,
	&g_sweep2_fx,
	&g_sweep3_fx,
	&g_sweep4_fx,
	&g_sweep5_fx,
	&g_harm_mask,
	&g_auto_harm_mask,
	&g_auto_subharm1_mask,
	&g_auto_subharm2_mask,
	&g_auto_subharm3_mask,
	&g_thresh_mask,
#ifdef STEREO
	&g_vocode,
	&g_harmmatchLR_fx,
	&g_harmmatchRL_fx
#endif
};

const int g_num_fft_fx = NUM_ELEMENTS(g_fft_fx_table);

FFTFx* GetFFTFx(int idx)
{
	if(idx < 0 || idx >= g_num_fft_fx) return &g_no_fx;
	return g_fft_fx_table[idx];
}
