/**************************************************************************************************
Copyright (c) 2005-2006 Darrell Tam
email: darrell.barrell@gmail.com

Each DtBlkFx effect has an FxCtrl gui panel
It contains the following controls:

Sliders for FreqA & FreqB that are aligned to the spectrograms in the main GUI
Amplitude invisi-slider
Effect type pull down menu
Effect value slider


History
	Date       Version     Programmer       Comments
	01/12/05    1.0        Darrell Tam		Created


This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

***************************************************************************************************/
#include "DtBlkFx.hpp"
#include "BlkFxParam.h"
#include "FFTFx.h"
#include "Gui.h"
#include "FxCtrl.h"
#include "Spectrogram.h"

#define DBG_FILE_NAME "c:\\fxctrl.txt"
#include "Debug.h"

// get blk fx out of the main gui class
inline DtBlkFx* FxCtrl::blkFx() { return _gui->blkFx(); }

inline void FxCtrl::sendParam(int param, float v)
{
	DBG("sendParam" << VAR(_param_offset) << VAR(param) << VAR(v));
	blkFx()->setParameterAutomated(param+_param_offset, v);
}

//-------------------------------------------------------------------------------------------------
FxCtrl::FxCtrl() :
	CViewContainer(
		CRect(0,0,0,0),		//const CRect &rect,
		NULL,				//CFrame *pParent,
		NULL				//CBitmap *pBackground
	)
{
	_ok = false;
}

//-------------------------------------------------------------------------------------------------
bool/*success*/FxCtrl::init(Gui* gui, CRect rect)
//
// expect gui to have been init'd with:
//   spectrograms
//   frame
{
	int i;
	_ok = false;
	_freq_sgram_vline = -1;

	// figure out where we are in the table
	_fx_set = this - gui->_fx_ctrl;
	_param_offset = BlkFxParam::paramOffs(_fx_set);
	DBG("FxCtrl::init" << VAR(this) << VAR(rect) << VAR(_param_offset));

	/*CView::*/size = rect;
	_mouse.top_left = TopLeft(rect);

	_gui = gui;
	setParent(gui->getFrame());

	// note, co-ordinates are relative to the top left of this control
	CPoint p(0, 2);
	CRect r(0, 0, 0, (rect.height()-5)/3);

	// place freq ctrls to align with spectrogram
	r.right = gui->_pix_freq.numPixels();
	p.x = gui->_sgram[0].client_rect.left-rect.left;

	for(i = 0; i < 2; i++) {
		_freq[i].initSlideRect(
			Offset(p, r),		// slider rectangle
			20					// knob width
		);
		DBG("freq" << VAR(r) << VAR(p) << VAR(_freq[i].slide_min) << VAR(_freq[i].slide_width));
		p.y += r.height();
	}

	// place amp ctrl
	p.x = 0;
	r.right = 100;
	_amp.init(
		Offset(p, r),			// rectangle
		100,					// pixels over which mouse must move for 0 to 1
		10						// number of pixels in y direction for 1/2 speed
	);
	
	p.x = 100;
	r.right = 100;
	// place fx type
	_fxtype_ctrl = 
		new COptionMenu (
			Offset(p, r),		// rectangle
			this,				// CControlListener *listener
			-1,					// tag number
			NULL,				// background
			NULL,				// background on click
			k3DIn				// style
		);
	if(!_fxtype_ctrl) return false;
	/*CViewContainer::*/addView(_fxtype_ctrl);

	for (i = 0; i < g_num_fft_fx; i++)
		_fxtype_ctrl->addEntry ((char*)(GetFFTFx(i)->name()), i);

	// place fx value
	p.x = 205;
	r.right = 200;
	_fxval.initSlideRect(
		Offset(p, r),			// slider rectangle
		20						// knob width
	);
	_fxtype_focus = false;

	_ok = true;
	DBG("FxCtrl" << VAR(_ok));
	return true;
}

//-------------------------------------------------------------------------------------------------
FxCtrl::~FxCtrl()
{
}

//-------------------------------------------------------------------------------------------------
void FxCtrl::drawRect(CDrawContext *context, CRect& update_rect)
// virtual, override CViewContainer
{
	DBG("drawRect" << VAR(this) << VAR(update_rect));

	// bodgey fix for change in params but only partial redraw - wait for the next update to redraw all
	// TODO: redraw everything (outside update_rect)
	if(_isDirty() && !IsContained(update_rect, size)) return;

	// get the effect type that we are drawing for
	int fx_type = (int)_fxtype_ctrl->getValue();
	FFTFx* fft_fx = GetFFTFx(fx_type);

	// create a temporary context for selecting bitmaps into
	CPoint top_left(0, 2);

	CColorRGB focus_txt_col(0xff,0x80,0x80), normal_txt_col(0xff,0xff,0xff);

	int i;
	CRect r;

	// copy background bitmap into our offscreen
	CBitmap* bg = pParent->getBackground();
	if(bg) bg->draw(pOffscreenContext, /*dst pos*/MoveToOrigin(size), /*src pos*/TopLeft(size));

	// copy on the amp overlay
	float amp_v = _amp.value;
	CAlphaContext*  amp_bm = _gui->_amp_bm;
	if(amp_bm && amp_bm->ok() && fft_fx->paramUsed(BlkFxParam::FX_AMP)) {
		CRect r = GetRect(amp_bm);
		r.right = (int)(amp_v*(float)r.right);
		if(r.width() > 0)
			amp_bm->blend(
				pOffscreenContext,				// dst
				Offset(top_left, r),			// dst rect
				r,								// src rect
				70								// overall alpha
			);
	}

	// draw the frequency overlay in 3 parts
	// if freqa < freqb, highlight the section between freq's
	// otherwise highlight sections outside those freq's
	//bm = _gui->_freq_bg_bm;
	CAlphaContext*  freq_bg = _gui->_freq_bg_bm;
	if(freq_bg && freq_bg->ok()) {
		//COffscreenContext bm_context(pOffscreenContext, bm);
		CRect r[3];
		for(i = 0; i < 3; i++) r[i] = GetRect(freq_bg);
		for(i = 0; i < 2; i++) r[i].right = _freq[i].knob_cent;

		// default to highlighting between freq's
		unsigned char alpha[3] = { 30, 120, 30 };

		// highlight outside freq's if freqa > freqb
		if(r[0].right > r[1].right) {
			std::swap(r[0].right, r[1].right);
			alpha[0] = 120;
			alpha[1] = 30;
			alpha[2] = 120;			
		}

		// draw main display
		r[1].left = r[0].right;
		r[2].left = r[1].right;
		for(int i = 0; i < 3; i++) {
			if(r[i].width() > 0)
				freq_bg->blend(
					pOffscreenContext,				// dst
					Offset(top_left, r[i]),			// dst rect
					r[i],							// src rect
					alpha[i]						// overall alpha
				);
		}

	}

	//
	// draw freq ctrl knobs
	//
	CColor freq_txt_col[2] = { focus_txt_col, focus_txt_col };
	for(i = 0; i < 2; i++) {
		CAlphaContext* freq_knob = _gui->_freq_knob_bm;
		if(!freq_knob || !freq_knob->ok()) continue;

		// expect bitmap to contain 3 images across (normal, focus, dragging) & 2 images high (freqA, freqB)
		CRect knob_r = GetRect(freq_knob);
		knob_r.bottom /= 2;
		knob_r.right /= 3;

		knob_r.offset(0, knob_r.bottom*i);

		// knob graphic depends on knob state
		switch(_freq[i].state) {
			case _freq[i].OVER_KNOB:
			case _freq[i].DRAGGING:
			case _freq[i].STEPPING:
				knob_r.offset(knob_r.right*2, 0); break;
			case _freq[i].FOCUS:
				knob_r.offset(knob_r.right, 0); break;
			default:
				freq_txt_col[i] = normal_txt_col;
		}

		r = _freq[i].getKnobRect(knob_r);
		// move knob down to bottom of control
		DBG("drawing freq knob" << VAR(r) << VAR(knob_r));
		unsigned char overall_alpha = 255;
		if(!fft_fx->paramUsed(BlkFxParam::FX_FREQ_A+i)) overall_alpha = 128;
		freq_knob->blend(/*dest*/pOffscreenContext, /*dest rect*/r, /*src rect*/knob_r, overall_alpha);
	}

	CColor fxval_txt_col = focus_txt_col;
	CAlphaContext* fxval_knob = _gui->_fxval_knob_bm;
	if(fxval_knob && fxval_knob->ok()) {
		// expect bitmap to contain 3 images (normal, focus, dragging)
		CRect knob_r = GetRect(fxval_knob);
		knob_r.right /= 3;

		// knob graphic depends on knob state
		switch(_fxval.state) {
			case _fxval.OVER_KNOB:
			case _fxval.DRAGGING:
			case _fxval.STEPPING:
				knob_r.offset(knob_r.right*2, 0); break;
			case _fxval.FOCUS:
				knob_r.offset(knob_r.right, 0); break;
			default:
				fxval_txt_col = normal_txt_col;
		}

		r = _fxval.getKnobRect(knob_r);
		r.offset(0, 3);
		DBG("drawing fxval knob" << VAR(r) << VAR(knob_r));
		unsigned char overall_alpha = 255;
		if(!fft_fx->paramUsed(BlkFxParam::FX_VAL)) overall_alpha = 128;
		fxval_knob->blend(/*dest*/pOffscreenContext, /*dest rect*/r, /*src rect*/knob_r, overall_alpha);
	}

	//
	// write some text on top of everything
	//
	pOffscreenContext->setFont(kNormalFont, /*sz*/0, kBoldFace);
	pOffscreenContext->setFillColor(CColorRGB(0,0,0));

	DBG("draw" << VAR(normal_txt_col) << VAR(focus_txt_col) << VAR(freq_txt_col[0]) << VAR(freq_txt_col[1]));
	// freq text
	char str[256], *s;
	for(i = 0; i < 2; i++) {
		pOffscreenContext->setFontColor(freq_txt_col[i]);

		float hz = BlkFxParam::getHz(_freq[i].value);

		r = _freq[i].getSlideRect();
		if(_freq[i].value<0.5f) r = Offset(TopRight(r), CRect(-100, 0, 0, r.height()));
		else r = Offset(TopLeft(r), CRect(0, 0, 100, r.height()));
		
		if(!fft_fx->paramUsed(BlkFxParam::FX_FREQ_A+i)) {
			s = str;
			s += sprintf(s, "not used");
			OutlineDrawString(
				pOffscreenContext,	// draw context
				r,					// rectangle
				str,				// string
				kCenterText			// horiz alignment
			);
		}
		else {
			s = str;
			s += HzToNote(s, hz);
			OutlineDrawString(
				pOffscreenContext,	// draw context
				LeftHalf(r),		// rectangle
				str,				// string
				kCenterText			// horiz alignment
			);

			s = str;
			s += sprintNum(s, hz, /*min unit*/1.0f);
			s += sprintf(s, "Hz");
			OutlineDrawString(
				pOffscreenContext,	// draw context
				RightHalf(r),		// rectangle
				str,				// string
				kCenterText			// horiz alignment
			);
		}
	}

	// amp text
	pOffscreenContext->setFontColor(
		_amp.state==_amp.INACTIVE ? normal_txt_col:focus_txt_col
	);

	s = str;
	bool mix_mode = fft_fx->ampMixMode();
	float amp_mult = BlkFxParam::getEffectAmpMult(amp_v, mix_mode);

	if(!fft_fx->paramUsed(BlkFxParam::FX_AMP))
		s += sprintf(s, "not used");
	else if(mix_mode && amp_mult <= 1.0f)
		s += sprintf(s, "%.0f %%", amp_mult*100.0f);
	else if(amp_mult > 0)
		s += sprintf(s, "%.0f dB", BlkFxParam::getEffectAmp(amp_v));
	else
		s += sprintf(s, "-inf dB");

	OutlineDrawString(
		pOffscreenContext,		// draw context
		GetRect(_amp),			// rectangle
		str,					// string, length
		kCenterText				// horiz alignment
	);
	
	// effect type text
	pOffscreenContext->setFontColor(_fxtype_focus?focus_txt_col:normal_txt_col);
	const char* name_str = fft_fx->name();
	OutlineDrawString(
		pOffscreenContext,		// draw context
		GetRect(_fxtype_ctrl),	// rectangle
		name_str,				// string, length
		kCenterText				// horiz alignment
	);

	// effect value text
	pOffscreenContext->setFontColor(
		_fxval.state!=_fxval.INACTIVE?focus_txt_col:normal_txt_col
	);
	s = str;
	s += fft_fx->dispVal(blkFx(), s, _fxval.value, _fx_set);
	OutlineDrawString(
		pOffscreenContext,		// draw context
		_fxval.getSlideRect(),	// rectangle
		str,					// string, length
		_fxval.value<0.5f?kRightText:kLeftText	// horiz alignment
	);

	// copy into main drawing area
	pOffscreenContext->copyFrom(context, /*dest*/size);

	// all redrawn
	_freq[0].setDirty(false);
	_freq[1].setDirty(false);
	_fxval.setDirty(false);
	_amp.setDirty(false);
	_fxtype_ctrl->setDirty(false);
	setDirty(false);
}

//-------------------------------------------------------------------------------------------------
void FxCtrl::draw(CDrawContext *context)
// virtual, override CViewContainer
// draw all
{
	drawRect(context, size);
}

//-------------------------------------------------------------------------------------------------
void FxCtrl::update(CDrawContext *context)
// virtual, override CViewContainer
//
// rely on being called regularly 
//
{
	_mouse.update(context);
	DBG("FxCtrl::update" << VAR(_fx_set) << VAR(_mouse.click_mask) << VAR(_mouse.pos) << VAR(_mouse.release_mask) << VAR(_freq[0].getKnobRect()));

	// send the event to all controls
	using namespace BlkFxParam;

	// clear lines unless either freq control is active
	bool clr_vline = true;

	// check for freq knob changes
	int i;
	for(i = 0; i < 2; i++) {
		if(_freq[i].update(_mouse)) sendParam(FX_FREQ_A+i, _freq[i].value);

		// claim vline if mouse is over control
		if(!_mouse.isAnyButtonDown() && _freq[i].isInside(_mouse))
			_freq_sgram_vline = i;

		// don't clear lines if control is active
		if(_freq[i].state != _freq[i].INACTIVE)
			clr_vline = false;
	}

	// place or remove lines
	if(_freq_sgram_vline != -1) {
		if(clr_vline) {
			for(i = 0; i < 2; i++)
				_gui->_sgram[i].clrVLine();
			_freq_sgram_vline = -1;
		}
		else {
			float frac = _freq[_freq_sgram_vline].value;
			for(i = 0; i < 2; i++)
				_gui->_sgram[i].setVLineFrac(frac);
		}
	}

	if(_fxval.update(_mouse)) sendParam(FX_VAL, _fxval.value);
	if(_amp.update(_mouse)) sendParam(FX_AMP, _amp.value);

	bool t = IsFocus(_fxtype_ctrl, _mouse);
	if(t != _fxtype_focus) {
		_fxtype_ctrl->setDirty(true);
		_fxtype_focus = t;
	}

	// check whether we need to redraw
	if(_isDirty()) draw(context);

	// don't pass the call down to CViewContainer (although it shouldn't do anything if we did)
}

//-------------------------------------------------------------------------------------------------
bool FxCtrl::_isDirty()
// internal
//
// check whether any control has moved and we need to redraw
//
{
	return
		_freq[0].isDirty() ||
		_freq[1].isDirty() ||
		_fxval.isDirty() ||
		_amp.isDirty() ||
		CViewContainer::isDirty();
}

//-------------------------------------------------------------------------------------------------
void FxCtrl::mouse(CDrawContext *context, CPoint &where)
// virtual, override CViewContainer
//
// mouse click
{
	_mouse.click(context, where, size);
	DBG("FxCtrl::mouse" << VAR(_fx_set) << VAR(_mouse.click_mask) << VAR(_mouse.pos) << VAR(_mouse.release_mask) << VAR(_freq[0].getKnobRect()));

	getParent()->setEditView(0);

	// send the click to controls to see if any will claim it

	// rely on "update()" to be called regularly to get mouse movement
	if(_mouse.click_mask) {
		// freq controls don't do default click behaviour
		int i;

		// modified behaviour: if the rt button is down but the mouse is not on the
		// knob then drag both controls
		for(i = 0; i < 2; i++) {
			if(_mouse.click_mask&kRButton &&
				!_freq[i].isInKnob(_mouse) &&
				_freq[i].isInside(_mouse)) {
				_freq_sgram_vline = i;
				_freq[0].forceState(_mouse, _freq[0].DRAGGING);
				_freq[1].forceState(_mouse, _freq[1].DRAGGING);
				return;
			}
		}

		// otherwise use the default click behaviour
		for(i = 0; i < 2; i++) {
			if(_freq[i].click(_mouse)) {
				_freq_sgram_vline = i;
				return;
			}
		}

		if(_fxval.click(_mouse)) return;
		if(_amp.click(_mouse)) return;
	}

	// otherwise check the ordinary vst gui control
	if(_fxtype_ctrl->hitTest(_mouse, _mouse.buttons)) {
		_fxtype_ctrl->mouse(context, /*CPoint*/_mouse);
		return;
	}
}

//-------------------------------------------------------------------------------------------------
long FxCtrl::onKeyDown(VstKeyCode& key)
{
	return CViewContainer::onKeyDown(key);
}

//-------------------------------------------------------------------------------------------------
long FxCtrl::onKeyUp (VstKeyCode& key)
{
	return CViewContainer::onKeyUp(key);
}

//-------------------------------------------------------------------------------------------------
bool FxCtrl::onWheel(CDrawContext *context, const CPoint &where, float distance)
{
	DBG("onWheel" << VAR(distance));
	CPoint p = where-TopLeft(size);
	_freq[0].changeIfFocus(p, distance);
	_freq[1].changeIfFocus(p, distance);
	_fxval.changeIfFocus(p, distance);

	return CViewContainer::onWheel(context, where, distance);
}

//-------------------------------------------------------------------------------------------------
void FxCtrl::takeFocus(CDrawContext *context)
// virtual, override CViewContainer
{
	CViewContainer::takeFocus(context);
}

//-------------------------------------------------------------------------------------------------
void FxCtrl::looseFocus(CDrawContext *context)
// virtual, override CViewContainer
{
	CViewContainer::looseFocus(context);
}

//-------------------------------------------------------------------------------------------------
bool FxCtrl::removed(CView* parent)
// virtual, override CViewContainer
{
	DBG("removed");
	CViewContainer::removed(parent);

	setParent(NULL);
	_ok = false;

	return true;
}
//-------------------------------------------------------------------------------------------------
void FxCtrl::valueChanged (CDrawContext *context, CControl *control)
// virtual, override CControlListener
//
// called by the gui control when changed by the user
{
	DBG("valueChanged" << VAR(_param_offset) << VAR(control));
	// check which control hit
	float v = control->getValue();
	using namespace BlkFxParam;
	if(control == _fxtype_ctrl) sendParam(FX_TYPE, getEffectTypeInv((long)v));

}

//-------------------------------------------------------------------------------------------------
void FxCtrl::setParameter(long index, float v)
//
// called when host application changes a parameter
//
// NOTE, we expect this to be called synchronously with "init" since we're relying on controls not
// being deleted
//
{
	DBG("setParameter" << VAR(_param_offset) << VAR(index) << VAR(v));
	if(!_ok) return;

	index -= _param_offset;
	using namespace BlkFxParam;
	switch(index) {
		case FX_FREQ_A: _freq[0].setValue(v); break;
		case FX_FREQ_B: _freq[1].setValue(v); break;
		case FX_AMP: _amp.setValue(v); break;
		case FX_TYPE: if(_fxtype_ctrl) _fxtype_ctrl->setValue((float)getEffectType(v)); break;
		case FX_VAL: _fxval.setValue(v); break;
	}
	// rely on update() to redraw if required
}
