/**************************************************************************************************
Copyright (c) 2005-2006 Darrell Tam
email: darrell.barrell@gmail.com

GlobalCtrl is the GUI panel at the top of the DtBlkFx window that lets you change the following
MixBack : % mixback of original audio (Owner invisi horiz slider)
Delay: Number of beats delay (Owner invisi horiz slider)
Overlap: % overlap of FFT blks (Owner invisi horiz slider)
Blk Sync: whether blks are sync'd to tempo ((Owner toggle)
Blk Len: pull down menu of FFT blk sz

History
	Date       Version    Programmer         Comments
	16/2/03    1.0        Darrell Tam		Created


This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

***************************************************************************************************/

#include "DtBlkFx.hpp"
#include "BlkFxParam.h"

#include "Gui.h"
#include "GlobalCtrl.h"

#define DBG_FILE_NAME "c:\\GlobalCtrl.txt"
#include "Debug.h"

#define DBG_PARAM_SAVE

// makesure real release doesn't do dbg param saving
#ifndef _DEBUG
#undef DBG_PARAM_SAVE
#endif

#ifdef DBG_PARAM_SAVE
// turn the mixback control into a save param on click instead
#include <fstream>
using namespace std;
#endif

static float g_delay_menu_entries[] = {
	0.25f,
	0.5f,
	0.75f,
	1.0f,
	1.25f,
	1.5f,
	1.75f,
	2.0f,
	2.5f,
	3.0f,
	3.5f,
	4.0f
};
enum { NUM_DELAY_MENU_ENTRIES = sizeof(g_delay_menu_entries)/sizeof(g_delay_menu_entries[0]) };

// get blk fx out of the main gui class
inline DtBlkFx* GlobalCtrl::blkFx()
	{ return _gui->blkFx(); }

inline void GlobalCtrl::sendParam(int param, float v)
{
	DBG("sendParam" << VAR(param) << VAR(v));
	blkFx()->setParameterAutomated(param, v);
}

//-------------------------------------------------------------------------------------------------
GlobalCtrl::GlobalCtrl() :
	CViewContainer(
		CRect(0,0,0,0),		//const CRect &rect,
		NULL,				//CFrame *pParent,
		NULL				//CBitmap *pBackground
	)
{
	_ok = false;
}

//-------------------------------------------------------------------------------------------------
bool/*success*/GlobalCtrl::init(Gui* gui, CRect rect)
//
// expect gui to have been init'd with:
//   spectrograms
//   frame
{
	int i;

	/*CView::*/size = rect;
	_mouse.top_left = TopLeft(rect);

	_ok = false;
	_gui = gui;
	setParent(gui->getFrame());

	_text_height = 15;

	// note, co-ordinates are relative to the top left of this control
	CPoint p(0, 0);
	CRect r(0, 0, rect.width()/6, rect.height());

	_mixback.init(
		Offset(p, r),			// rectangle
		100,					// pixels over which mouse must move for 0 to 1
		10						// number of pixels in y direction for 2x speed
	);	
	p.x += r.width();
	_pwrmatch.init(
		Offset(p, r)			// rectangle
	);
	p.x += r.width();
	// delay has both an invisi-slider and a drop down menu
	_delay.init(
		Offset(p, r),			// rectangle
		100,					// pixels over which mouse must move for 0 to 1
		10						// number of pixels in y direction for 2x speed
	);
	_delay_menu =
		new COptionMenu (
			Offset(p, r),		// rectangle
			this,				// CControlListener *listener
			-1,					// tag number
			NULL,				// background
			NULL,				// background on click
			k3DIn				// style
		);

	if(!_delay_menu) return false;
	/*CViewContainer::*/addView(_delay_menu);
	for (i = 0; i < NUM_DELAY_MENU_ENTRIES; i++) {
		char str[128];
		sprintf(str, "%.2f beats", g_delay_menu_entries[i]);
		_delay_menu->addEntry(str, i);
	}

	p.x += r.width();
	_overlap.init(
		Offset(p, r),			// rectangle
		100,					// pixels over which mouse must move for 0 to 1
		10						// number of pixels in y direction for 2x speed
	);
	p.x += r.width();

	_blksync.init(
		Offset(p, r)			// rectangle
	);
	p.x += r.width();

	_blksz =
		new COptionMenu (
			Offset(p, r),		// rectangle
			this,				// CControlListener *listener
			-1,					// tag number
			NULL,				// background
			NULL,				// background on click
			k3DIn				// style
		);

	if(!_blksz) return false;

	/*CViewContainer::*/addView(_blksz);
	for (i = 0; i < N_FFT_BLK_SZ; i++) {
		char str[128];
		char* s = str;
		// fft length 
		int fft_len = GET_FFT_BLK_SZ(i);

		// seconds fft length corresponds to
		float secs = (float)fft_len/blkFx()->getSampleRate();
		s += sprintNum(s, secs);
		s += sprintf(s, "sec (%d samps)", fft_len);
		_blksz->addEntry (str, i);
		_delay_menu->addEntry(str, NUM_DELAY_MENU_ENTRIES+i);
	}

	_blksz_focus = false;
	_ok = true;
	return true;
}

//-------------------------------------------------------------------------------------------------
GlobalCtrl::~GlobalCtrl()
{
}

//-------------------------------------------------------------------------------------------------
void GlobalCtrl::drawRect(CDrawContext *context, CRect& update_rect)
// virtual, override CViewContainer
{
	DBG("drawRect" << VAR(this) << VAR(update_rect));

	// bodgey fix for change in params but only partial redraw - wait for the next update to redraw all
	// TODO: redraw everything (outside update_rect)
	if(_isDirty() && !IsContained(update_rect, size)) return;

	CColorRGB focus_txt_col(0xff,0x80,0x80), normal_txt_col(0xff,0xff,0xff);

	// copy background bitmap into our offscreen
	CBitmap* bg = pParent->getBackground();
	if(bg) bg->draw(pOffscreenContext, /*dst pos*/MoveToOrigin(size), /*src pos*/TopLeft(size));

	//
	// write some text on top of everything
	//
	pOffscreenContext->setFont(kNormalFont, /*sz*/0, kBoldFace);
	pOffscreenContext->setFillColor(CColorRGB(0,0,0));

	using namespace BlkFxParam;

	char str[256];
	char* s;

	// mixback text
	pOffscreenContext->setFontColor(_mixback.state==_mixback.INACTIVE ? normal_txt_col:focus_txt_col);
	s = str;
	s += sprintf(s, "%.0f%%", _mixback.value*100.0f);
	OutlineDrawString(
		pOffscreenContext,		// draw context
		textRect(GetRect(_mixback)),		// rectangle
		str,					// string, length
		kCenterText				// horiz alignment
	);

	// power match
	pOffscreenContext->setFontColor(_pwrmatch.state==_pwrmatch.INACTIVE ? normal_txt_col:focus_txt_col);
	s = str;
	s += sprintf(s, _pwrmatch.pending_value?"match":"filter");
	OutlineDrawString(
		pOffscreenContext,		// draw context
		textRect(GetRect(_pwrmatch)),			// rectangle
		str,					// string, length
		kCenterText				// horiz alignment
	);

	// delay
	pOffscreenContext->setFontColor(_delay.state==_delay.INACTIVE ? normal_txt_col:focus_txt_col);
	float delay_samps = (float)(blkFx()->getDelaySamps(_delay.value)-blkFx()->_initial_delay);
	float delay_beats = delay_samps/blkFx()->getSampsPerBeat();
	float delay_secs = delay_samps/blkFx()->getSampleRate();
	s = str;
	s += sprintf(s, "%.2f beats", delay_beats);
	OutlineDrawString(
		pOffscreenContext,		// draw context
		textRect(GetRect(_delay)),		// rectangle
		str,					// string, length
		kCenterText				// horiz alignment
	);

	// overlap text
	pOffscreenContext->setFontColor(_overlap.state==_overlap.INACTIVE ? normal_txt_col:focus_txt_col);
	s = str;
	float guess_fft_len = blkFx()->guessFFTLen();
	float samps_fwd = (float)BlkFxParam::getBlkShiftFwd(
		BlkFxParam::getOverlapParam(_overlap.value, _blksync.pending_value),
		guess_fft_len
	);
	float overlap_frac = 1.0f-(samps_fwd/guess_fft_len);
	s += sprintf(s, "%.0f%%", overlap_frac*100.0f);
	OutlineDrawString(
		pOffscreenContext,		// draw context
		textRect(GetRect(_overlap)),		// rectangle
		str,					// string, length
		kCenterText				// horiz alignment
	);

	// blk sync
	pOffscreenContext->setFontColor(_blksync.state==_blksync.INACTIVE ? normal_txt_col:focus_txt_col);
	s = str;
	s += sprintf(s, _blksync.pending_value?"on":"off");
	OutlineDrawString(
		pOffscreenContext,		// draw context
		textRect(GetRect(_blksync)),			// rectangle
		str,					// string, length
		kCenterText				// horiz alignment
	);

	// fft len
	pOffscreenContext->setFontColor(_blksz_focus?focus_txt_col:normal_txt_col);
	int fft_plan = limit_range((int)_blksz->getValue(), 0, N_FFT_BLK_SZ-1);
	int fft_len_set = GET_FFT_BLK_SZ(fft_plan);
	float fft_len = blkFx()->guessFFTLen();
	float fft_secs = fft_len/blkFx()->getSampleRate();
	s = str;
	s += sprintNum(s, fft_secs);
	s += sprintf(s, "sec");

	// put an asterisk after if the FFT len is longer than the delay
	if(delay_samps < fft_len) s += sprintf(s, " *");
	OutlineDrawString(
		pOffscreenContext,		// draw context
		textRect(GetRect(_blksz)),		// rectangle
		str,					// string, length
		kCenterText				// horiz alignment
	);

	// copy into main drawing area
	pOffscreenContext->copyFrom(context, size);

	// all redrawn
	_mixback.setDirty(false);
	_pwrmatch.setDirty(false);
	_delay.setDirty(false);
	_overlap.setDirty(false);
	_blksync.setDirty(false);
	_blksz->setDirty(false);
	setDirty(false);
}

//-------------------------------------------------------------------------------------------------
void GlobalCtrl::draw(CDrawContext *context)
// virtual, override CViewContainer
// draw all
{
	drawRect(context, size);
}

//-------------------------------------------------------------------------------------------------
void GlobalCtrl::update(CDrawContext *context)
// virtual, override CViewContainer
//
// rely on being called regularly 
//
{
	_mouse.update(context);

	// send the event to all controls
	using namespace BlkFxParam;

	bool mixback_update = _mixback.update(_mouse);
	if(_pwrmatch.update(_mouse)) mixback_update = true;
	if(mixback_update) sendParam(MIX_BACK, getMixbackParam(_mixback.value, _pwrmatch.getValue()));

	if(_pwrmatch.update(_mouse)) sendParam(MIX_BACK, _mixback.value);

	if(_delay.update(_mouse)) sendParam(DELAY, _delay.value);
	
	// combine the overlap & sync controls into a single overlap param
	bool overlap_update = _overlap.update(_mouse);
	if(_blksync.update(_mouse)) overlap_update = true;
	if(overlap_update) sendParam(OVERLAP, getOverlapParam(_overlap.value, _blksync.getValue()));

	// chk whether blk sz has focus
	bool t = IsFocus(_blksz, _mouse);
	if(t != _blksz_focus) {
		_blksz->setDirty(true);
		_blksz_focus = t;
	}

	// check whether we need to redraw
	if(_isDirty()) draw(context);

	// don't pass the call down to CViewContainer (although it shouldn't do anything if we did)
}

//-------------------------------------------------------------------------------------------------
bool GlobalCtrl::_isDirty()
// internal
//
// check whether any control has moved and we need to redraw
{
	return
		_mixback.isDirty() ||
		_pwrmatch.isDirty() ||
		_delay.isDirty() ||
		_overlap.isDirty() ||
		_blksync.isDirty() ||
		CViewContainer::isDirty();
}

//-------------------------------------------------------------------------------------------------
void GlobalCtrl::mouse(CDrawContext *context, CPoint &where)
// virtual, override CViewContainer
//
// mouse click
{
	_mouse.click(context, where, size);

	//getParent()->setEditView(0);

	// send the click to controls to see if any will claim it

	// rely on "update()" to be called regularly to get mouse movement
	if(_mouse.click_mask) {
		if(_mixback.click(_mouse, /*r button scale*/0.25f)) {
			#ifdef DBG_PARAM_SAVE
			ofstream params_save("c:\\params_save.txt", ios_base::app);
			params_save << "{ \"\", { ";
			int i;
			for(i = 0; i < BlkFxParam::TOTAL_NUM-1; i++)
				params_save << "(float)" << blkFx()->getCurrParam(i) << ", ";
			params_save << "(float)" << blkFx()->getCurrParam(i) << "} }," << endl;
			#endif
			return;
		}
		if(_pwrmatch.click(_mouse)) return;

		// only pass right button to delay invisi-slider
		if(_mouse.click_mask&kRButton)
			if(_delay.click(_mouse, /*r button scale*/0.05f)) return;

		if(_overlap.click(_mouse, /*r button scale*/0.25f)) return;
		if(_blksync.click(_mouse)) return;
	}

	// pass the event down if we didn't take it
	CViewContainer::mouse(context, where);
}

//-------------------------------------------------------------------------------------------------
long GlobalCtrl::onKeyDown(VstKeyCode& key)
{
	return CViewContainer::onKeyDown(key);
}

//-------------------------------------------------------------------------------------------------
long GlobalCtrl::onKeyUp (VstKeyCode& key)
{
	return CViewContainer::onKeyUp(key);
}

//-------------------------------------------------------------------------------------------------
bool GlobalCtrl::onWheel(CDrawContext *context, const CPoint &where, float distance)
{
	DBG("onWheel" << VAR(distance));
	return CViewContainer::onWheel(context, where, distance);
}

//-------------------------------------------------------------------------------------------------
void GlobalCtrl::takeFocus(CDrawContext *context)
// virtual, override CViewContainer
{
	CViewContainer::takeFocus(context);
}

//-------------------------------------------------------------------------------------------------
void GlobalCtrl::looseFocus(CDrawContext *context)
// virtual, override CViewContainer
{
	CViewContainer::looseFocus(context);
}

//-------------------------------------------------------------------------------------------------
bool GlobalCtrl::removed(CView* parent)
// virtual, override CViewContainer
{
	DBG("removed");
	CViewContainer::removed(parent);

	setParent(NULL);
	_ok = false;
	// could cleanup more stuff

	return true;
}
//-------------------------------------------------------------------------------------------------
void GlobalCtrl::valueChanged (CDrawContext *context, CControl *control)
// virtual, override CControlListener
//
// called by the gui control when changed by the user (only for VstGui controls, not owner draw type)
//
{
	// check which control hit
	float v = control->getValue();

	DBG("valueChanged" << VAR(control) << VAR(v));
	if(control == _blksz) sendParam(BlkFxParam::FFT_LEN, BlkFxParam::getFFTLenParam((int)v));

	else if(control == _delay_menu) {
		float beats;
		if(v >= 0 && v < NUM_DELAY_MENU_ENTRIES)
			beats = g_delay_menu_entries[(int)v];
		else {
			// delay corresponds to a block size
			int delay_samps = GET_FFT_BLK_SZ((int)v-NUM_DELAY_MENU_ENTRIES);
			beats = (float)delay_samps / blkFx()->getSampsPerBeat();
		}
		float param = BlkFxParam::getDelayParam(beats);
		sendParam(BlkFxParam::DELAY, param);
		_delay.setValue(param);
	}
}

//-------------------------------------------------------------------------------------------------
void GlobalCtrl::setParameter(long index, float v)
//
// called when host application changes a parameter
//
// NOTE, we expect this to be called synchronously with "init" since we're relying on controls not
// being deleted
//
{
	DBG("setParameter" << VAR(index) << VAR(v) << VAR(_ok));
	if(!_ok) return;

	using namespace BlkFxParam;
	switch(index) {
		case MIX_BACK:
			_mixback.setValue(getMixBackFrac(v));
			_pwrmatch.setValue(getPwrMatch(v));
			break;

		case DELAY: _delay.setValue(v); break;

		case FFT_LEN:
			if(_blksz) _blksz->setValue((float)getPlan(v)); break;

		case OVERLAP: 
			_overlap.setValue(getOverlapPart(v));
			_blksync.setValue(getBlkSync(v));
			break;
	}
	// rely on update() to redraw if required
}
