/**************************************************************************************************
Copyright (c) 2005-2006 Darrell Tam
email: darrell.barrell@gmail.com

PixelFreqMap (and back) is used by spectrograms and DtBlkFx gui to display any frequency values
Frequencies are displayed on a log scale so as to be linear for music notes

This class uses the DtBlkFx freq param (in BlkFxParam) routines to generate the mapping

History
	Date       Version    Programmer         Comments
	01/12/05    1.0        Darrell Tam		Created


This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

***************************************************************************************************/
#include <dt/misc_stuff.h>
#include "BlkFxParam.h"
#include "PixelFreqMap.h"

#define DBG_FILE_NAME "c:\\PixelFreqMap.txt"
#include "Debug.h"

using namespace std;

//-------------------------------------------------------------------------------------------------
PixelFreqMap::PixelFreqMap(void) { }
PixelFreqMap::~PixelFreqMap(void) { }

//-------------------------------------------------------------------------------------------------
void PixelFreqMap::init(
	int n_pixels,
	float sample_freq			// Hz
)
// generate the pixel to frequency mapping
{
	_sample_freq = sample_freq;

	if(n_pixels < 2) n_pixels = 2;

	// octaves are displayed with constant spacing
	_pix_per_octave = (float)(n_pixels-1)*BlkFxParam::octaveFrac();

	// find what freq each pixel will correspond to
	_pix_to_freq.resize(n_pixels);
	float param = 0.0f;
	float param_step = 1.0f/(float)(n_pixels-1);
	for(int i = 0; i < n_pixels; i++) {
		// this is actually a linear-to-exp mapping that we could generate much faster but
		// calling the conversion guarantees a match with the vst param
		_pix_to_freq[i] = BlkFxParam::getHzNonZero(param);
		DBG(VAR(i) << VAR(_pix_to_freq[i]) << VAR(param));
		param += param_step;
	}

	// multiplier to get frequency shift for +0.5 pixel
}


//-------------------------------------------------------------------------------------------------
void PixelFreqMap::BinRng::init(
	PixelFreqMap& fm,
	float* fft_dat				// fft data as produced by fftw
)
// expect the freq map to have been init'd already
{
	// get the pixel to freq map
	float* pix_to_freq = fm.pixToFreq();

	float half_pix_mult = powf(2.0f, 0.5f/fm._pix_per_octave)/fm._sample_freq;

	// generate pixel to fft-bin for each fft length
	// each pixel corresponds to a range of 1 or more bins in the fft
	//
	int fft_len = FFT_BLK_SZ_0;
	for(int i = 0; i < N_FFT_BLK_SZ; i++) {
		map[i].resize(fm.numPixels()+1);

		// find scaling to turn _pix_to_freq[x+0.5] into fft bin position
		float bin_scale = half_pix_mult*(float)fft_len;

		int bin = 0;				// start at bin 0
		int max_bin = fft_len/2;	// max bin is 1/2 fft len

		map[i][0] = pcplxf(fft_dat, fft_len, 0);
		for(int x=1; x <= fm.numPixels(); x++) {
			// bin corresponds to pixel x-0.5, since x-1 index and bin_scale includes +0.5 pixel
			int bin = (int)(bin_scale*pix_to_freq[x-1] + 0.5f);
			if(bin > max_bin) bin = max_bin;

			// convert bin start and end into fft data pointers
			map[i][x] = pcplxf(fft_dat, fft_len, bin);
			pcplxf p(fft_dat, fft_len, bin);
		}
		fft_len <<= 1;
	}
}

