#ifndef _DT_SPECTROGRAM_H_
#define _DT_SPECTROGRAM_H_
/**************************************************************************************************
Copyright (c) 2005-2006 Darrell Tam
email: darrell.barrell@gmail.com

Pls see Spectrogram.cpp

History
	Date       Version    Programmer         Comments
	16/2/03    -        Darrell Tam		Buzz version
    01/12/05   1.0        DT                 ported to VST

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

***************************************************************************************************/

#include <dt/misc_stuff.h>
#include <dt/VstGuiSupport.h>

#include "PixelFreqMap.h"
#include "GradColMap.h"


//------------------------------------------------------------------------------------------
class Spectrogram : public CViewContainer
{
public: // initialization routines that MUST be called before use

	Spectrogram();
	virtual ~Spectrogram();

	// set the colour mapping and 
	void set(GradColMap& /*in*/ col_map);

	// set the frequency-pixel mapping & initialize (must call before use)
	void init(CFrame* parent, CPoint p, int height, PixelFreqMap* pix_freq);

	// get image - to do: could un-circular buf image so that caller can draw into it
	CAlphaContext* getImage() { return _image; }

	// call when main effect suspends
	void suspend();

	// only call newData & blkDone if rdy() returns true
	bool rdy() { return _ok && !paused; }

	// supply new FFT data 
	void newData(
		pcplxf* bin_rng,	// appropriate bin rng's from PixelFreqMap
		float pwr_scale		// scaling to apply to data
	);

	// call after "newData()" has been called for each channel to update sample position
	void blkDone(
		long samp_pos,		// absolute sample position for start of this blk
		long time_fft_n		// number of samples in time domain (may be less samples than fft freq data)
	);

	// set vertical line using pixel position (as an offset into the client rect)
	void setVLinePix(int x) { _vline_x[1] = limit_range(x, 0, _image->getWidth()-1); }

	// set vertical line using a fraction
	void setVLineFrac(float/*0..1*/v) { setVLinePix((int)(v*(float)_image->getWidth())); }

	// clear the vertical line
	void clrVLine() { _vline_x[1] = -1; }

	// height of octave ticks at bottom in pixels
	int tick_height;

	// frame & tick colour
	CColor frame_col;

	// minimum number of samples per line of display (to limit scrolling speed)
	int min_samps_per_line;

	// client area (display area within borders & octave ticks)
	CRect client_rect;

	// text rectangle to display mouse position info
	CRect info_rect;

	// rectange to display menu text
	CRect menu_rect;

	// whether or not we're paused
	bool paused;

	// owner menu
	VstGuiRef<COptionMenu> menu;

	// name displayed if menu is not set
	const char* name;

public:	// override CViewContainer methods
	virtual void draw(CDrawContext *context);
	virtual void drawRect(CDrawContext *pContext, CRect& update_rect);
	virtual bool isDirty() { return true; } // update() is always called
	virtual void update(CDrawContext *context);
	virtual void mouse(CDrawContext *context, CPoint &where);
	virtual bool removed(CView* parent);

protected: // internals
	void invertVLine(CDrawContext *context);
	void eraseOverlay(CDrawContext *context);
	void drawOverlay(CDrawContext *context);

	unsigned long mapCol(float in);

	void getMaxVals(
		pcplxf* bin_rng,	// appropriate bin rng's from PixelFreqMap
		bool empty_line,	// whether this is a new line (pass as constant to help optimizer)
		float pwr_scale		// pwr scale to apply
	);

	void drawBm(
		CDrawContext *context,	//
		CRect clip_rect			// limit drawing to this clipping rectangle
	);

protected:
	// pixel to frequency mapping
	Ptr<PixelFreqMap> _pix_freq;

	// map to use for converting FFT bin pwr into a colour
	Ptr<GradColMap> _col_map;

	// colour scaling of log(pwr(fft data))
	float _col_log_offs, _col_log_scale;

	// scaling to use on FFT bin pwr when mapping to a colour
	float _pwr_min, _pwr_max;

	// 2D image data array
	// lines are arranged in a circular buffer
	VstGuiRef<CAlphaContext> _image;

	// current line position in image_data & _samp_pos
	int _image_i;

	// we can display up to this line
	int _disp_i[2];			// [0]=currently displayed, [1]=can display to

	// absolute sample position of each line in the image_data
	std::valarray<long> _line_samp_pos;

	// current max values from fft for each pixel
	std::valarray<float> _curr_max;

	// sample position of previously displayed line
	long _curr_line_samp_pos;

	// true if a line has just been output
	bool _curr_line_is_empty;

	// vertical line under mouse
	int _vline_x[2];		// [0]=current drawn, [1]=new value (-1=don't draw)

	// mouse info text state
	bool _info_drawn, _info_dirty;

	// menu text state
	bool _menu_drawn, _menu_dirty;

	// 
	MouseUpdate _mouse;

	// true when all was init'd ok
	bool _ok;

	// true when we have focus
	bool _focus;

};

#endif
