/******************************************************************************

Generic Parameter FIFO for VST params

History
	Date       Version    Programmer         Comments
	16/2/03    1.0        Darrell Tam		 Created
	14/2/04    1.1        DT                 Changed to template based
	9/11/05    1.2        DT                 Ported from buzz to vst

This is completely free software

******************************************************************************/

#ifndef _DT_PARAMS_DELAY_H_
#define _DT_PARAMS_DELAY_H_

#include <vector>
#include "misc_stuff.h"

//-------------------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------------------
template <int _N /*number of params*/ > class ParamsDelay
{

public:
	enum { N = _N };

	struct Params
	{
		// parameter values
		float vals[N];

		// whether or not each value was explicitly set
		bool explicit_set[N];

		// true if any of "explicit_set" is true
		bool any_explicit_set;

		// absolute sample position for these params
		long samp_abs;
	};

	typedef std::vector<Params> ParamsVec;
	typedef typename ParamsVec::iterator ParamsVecIterator;

	// circular buffer parameters
	ParamsVec _v;

	// current input position, NULL if there are no params in _v
	ParamsVecIterator _in;

	// current output position, NULL if thee is less than 2 params in _v
	ParamsVecIterator _out[2];

	// expected distance between samp_abs in Params (normally matches a fraction of the
	// tempo, e.g. samples per quarter-note)
	long _expected_dist;

	// current interpolation fraction of between _out[0] & _out[1] (set by setOutPos)
	float _interp_frac;

	void incPtr(ParamsVecIterator &p) { if(++p==_v.end()) p = _v.begin(); }

public:
	ParamsDelay()
	{
		_v.resize(128);	// arbitary initial size
		_expected_dist = 0x7fffffff; // large initial setting (disable)
		init();
	}

	void setExpectedDist(int n) { _expected_dist = n; }

	//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// use "setOutPos" to set the sample position

	// get linear interpolated value of current position
	float getInterp(int idx) const
		{ return lin_interp(_interp_frac, _out[0]->vals[idx], _out[1]->vals[idx]); }

	// get non-interpolated value of current param position rounded down (or up if next_param true)
	float getNonInterp(int idx, bool next_param=false) const { return _out[next_param?1:0]->vals[idx]; }

	// return true if "getNonInterp()" corresponds to an explicitly set parameter
	bool isExplicitlySet(int idx, bool next_param=false) const { return _out[next_param?1:0]->explicit_set[idx]; }

	// get the absolute sample position for the current param
	long getSampAbs(bool next_param=false) const { return _out[next_param?1:0]->samp_abs; }

	// return true if any param is explicitly set
	long isAnyExplicitSet(bool next_param=false) const { return _out[next_param?1:0]->any_explicit_set; }

	// get the most recent value at "idx"
	float getInput(int idx) const { return _in->vals[idx]; }

	//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	void init()
	{
		_in = _v.begin();
		_out[0] = _in;
		_out[1] = _in;
		_interp_frac = 0.0f;
		memset(&(*_in), 0, sizeof(Params));
	}

	//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	void resetAndCopyIn()
	// reset the params vector and copy the most recent params but reset abs time to 0
	{		
		Params in_params = *_in;
		init();
		*_in = in_params;
		_in->samp_abs = 0;
	}

	//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	bool/*already exists*/put(long samp_abs, int idx, float value)
	//
	// put a parameter with index "idx" and value "value" into a param node at absolute
	// sample position "samp_abs"
	//
	// return true if the param already exists
	{
		bool already_exists = false;
		// does this param have the same time as the most recent one?
		long samp_diff = samp_abs - _in->samp_abs;
		if(samp_diff <= 0) already_exists = true;
		else {
			// make a new Params entry if the time is different

			// make an extra entry point if the time gap is bigger than the expected
			// parameter distance
			if(samp_diff > _expected_dist) {
				cpInNode();
				_in->any_explicit_set = false;
				_in->samp_abs = samp_abs - _expected_dist;

			}
			cpInNode();
			_in->any_explicit_set = true;
			_in->samp_abs = samp_abs;
		}
		_in->vals[idx] = value;
		_in->explicit_set[idx] = true;
		return already_exists;
	}

	//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	bool/*can interpolate*/setOutPos(long samp_abs, bool find_fraction=true)
	//
	// set output sample position 
	// Note, only move forward using this
	//
	{
		// look for params 
		while(1) {

			if(_out[1]->samp_abs > samp_abs) {
				// find distance from requested sample to _out[0] position
				long numer = samp_abs - _out[0]->samp_abs;

				// can't go in reverse, samp_abs lies before "_out[0]"
				if(numer < 0) {
					_interp_frac = 0.0f;
					return false;
				}
				// can interpolate ok

				// don't do fraction calculation unless needed
				if(!find_fraction) return true;

				// distance between params
				float denom = (float)(_out[1]->samp_abs - _out[0]->samp_abs);

				// denominator should never be <=0, but just in case
				if(denom <= 0.0f) {
					_interp_frac = 0.0f;
					return true;
				}

				_interp_frac = (float)numer/denom;
				return true;
			}

			// is it an exact match?
			if(_out[1]->samp_abs == samp_abs) {
				_out[0] = _out[1];
				if(_out[1] != _in) incPtr(_out[1]);
				_interp_frac = 0.0f;
				return true;
			}

			// have we hit the end? Put both out ptrs at the end
			if(_out[1] == _in) {
				_out[0] = _out[1];
				_interp_frac = 0.0f;
				return false;
			}

			_out[0] = _out[1];
			incPtr(_out[1]);
		}
	}

protected:
	//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	void cpInNode()
	//
	// make a new node past current "_in" and that is a copy with all "explicit_set" to false
	//
	{
		ParamsVec::iterator prev = _in;
		incPtr(_in);

		// copy previous
		memcpy(_in->vals, prev->vals, sizeof(_in->vals));

		// clear explicit sets
		memset(_in->explicit_set, 0, sizeof(_in->explicit_set));

		// check whether we've filled the buffer
		if(_in == _out[0]) {
			// push the out pointers forward
			incPtr(_out[0]);
			incPtr(_out[1]);
		}
	}
};




#endif
