#ifndef _VST_OWNER_CTRLS_H_
#define _VST_OWNER_CTRLS_H_
/**************************************************************************************************
Copyright (c) 2005-2006 Darrell Tam
email: darrell.barrell@gmail.com

VstGuiOwnerControls.h
OwnerControls (the base class) are controls that don't draw their own graphics. The methods 
resemble the VST controls

The derived classes are:
OwnerToggle : toggle button

OwnerHSlider: normal horizontal slider with knob. Left button gives user normal control, right
    button gives fine control

OwnerInvisiHSlider: not really a slider but similar in that you click & drag the mouse left or
    right to change. Left button is normal change speed, right button is fine tune speed

OwnerKnob: Circular knob requiring user to move mouse in a circle around the knob

History
	Date       Version    Programmer         Comments
	01/12/05    1.0        Darrell Tam		Created


This is completely free software

***************************************************************************************************/

#include "VstGuiSupport.h"
#include <math.h>

//-------------------------------------------------------------------------------------------------
class OwnerControl
{
public:
	OwnerControl()
	{
		value = 0.0f;
		val_chg = true;
	}
	virtual ~OwnerControl() {}

	virtual bool/*value changed*/setValue(float v)
	// set the value and update knob_rect
	{
		v = limit_range(v, 0.0f, 1.0f);
		if(v == value) return false;
		value = v;
		val_chg = true;
		return true;
	}

	virtual bool isInside(CPoint p) { return full_rect.pointInside(p); }

	virtual bool/*value changed*/changeIfFocus(CPoint p, float delta)
	// only add "delta" to value if mouse is within our rectangle
	{
		if(!isInside(p)) return false;
		return setValue(value+delta);
	}

	// full rectangle containing control
	CRect full_rect;

	// current value (0.0f .. 1.0f)
	float value;

	// true if value changed after calling "setValue"
	bool val_chg;

};
inline CRect GetRect(const OwnerControl& c) { return c.full_rect; }


//-------------------------------------------------------------------------------------------------
class OwnerToggle : public OwnerControl
//
// owner toggle
// value toggles between 0 & 1
//
{
public:
	typedef enum {
		FORCE_DIRTY=-1,		// internal state used for marking as dirty
		INACTIVE=0,			// mouse not over rect
		FOCUS,				// mouse is over us
		GOT_CLICK			// we've got the click and we're waiting for the release before changing
	} State;

	void init(CRect full_rect_)
	{
		full_rect = full_rect_;
		pending_value = value?true:false;
		state = INACTIVE;
		setDirty(true);
	}

	bool/*true=click within rectangle*/click(const MouseUpdate& mouse)
	// call when there's a mouse click
	{
		if(!isInside(mouse)) return false;
		state = GOT_CLICK;
		// toggle
		pending_value = !getValue();
		return true;
	}

	bool/*value changed*/update(const MouseUpdate& mouse)
	//
	// return true if the value has changed
	// cancel change if mouse moves out of our rectangle
	{
		// go to inactive anytime mouse moves outside rectangle
		if(!isInside(mouse)) {
			if(state == GOT_CLICK) pending_value = getValue(); // cancel pending value
			state = INACTIVE;
			return false;
		}

		// still in our rectangle - look for button up
		if(state == GOT_CLICK) {
			if(!mouse.isAnyButtonDown(/*raw*/false)) {
				state = FOCUS;

				// apply change
				if(getValue() == pending_value) return false;
				value = pending_value?1.0f:0.0f;
				val_chg = true;
				return true;
			}
			return false;
		}

		// take focus as long as no button click outside our rectangle
		if(mouse.isAnyButtonDown(/*raw*/true)) state = INACTIVE;
		else state = FOCUS;
		return false;
	}


	// return true if knob needs redraw because value changed or state has changed
	bool /*true=needs redraw*/ isDirty() const
		{ return val_chg || state != state_drawn; }

	void setDirty(bool dirty)
	// set or clear dirty state
	{
		val_chg = dirty;
		state_drawn = dirty?FORCE_DIRTY:state;
	}

	bool getValue() const { return value?true:false; }

	virtual bool/*value changed*/setValue(float v)
	// override OwnerControl
	// set the value and update knob_rect
	{
		if(!OwnerControl::setValue(v)) return false;
		// update pending value unless user has clicked on us
		if(state != GOT_CLICK) pending_value = getValue();
		return true;
	}

	// during click, pending_value is the opposite to value - this is the value
	// that should be always drawn to the screen
	bool pending_value;

	// state
	State state, state_drawn;
};

//-------------------------------------------------------------------------------------------------
class OwnerHSlider : public OwnerControl
//
// horizontal slider support for classes that draw their own slider graphics
// value of 0.0f corresponds to knob at the left
//
{
public:
	// state of slider
	typedef enum {
		FORCE_DIRTY=-1,		// internal state used for marking as dirty
		INACTIVE=0,			// mouse not over slider
		FOCUS,				// mouse is inside slider rectangle (but not over knob)
		OVER_KNOB,			// mouse is over knob
		DRAGGING,			// mouse clicked on knob
		STEPPING			// mouse clicked in rectangle but not on knob
	} State;

	OwnerHSlider()
	{
		step_speed = 0.005f;
	}

	void _init()
	{
		state = INACTIVE;
		setDirty(true);
	}

	void initSlideRect(
		CRect slide_rect,				// full rectangle extends 1/2 knob_rect either side
		int knob_width_					// width of knob
	)
	// set the slide rectangle (rectangle bounding knob centre)
	{
		_init();
		slide_min = slide_rect.left;
		slide_width = (float)slide_rect.width();
		knob_width = knob_width_;
		full_rect = slide_rect;
		full_rect.left -= knob_width/2;
		full_rect.right += knob_width/2;
		setValue(value);
	}	

	void initFullRect(
		CRect full_rect_,				// slider rectangle is 1/2 knob width smaller either side
		int knob_width_					// width of knob
	)
	// set full rectangle (rectangle bounding full knob graphics)
	{
		_init();
		full_rect = full_rect_;
		knob_width = knob_width_;

		slide_min = full_rect.left+knob_width/2;
		slide_width = (float)(full_rect.width()-knob_width);
		setValue(value);
	}


	// return rectangle of knob sized "r" centred around knob centre
	CRect getKnobRect(CRect r) const
		{ return SetCent(CPoint(knob_cent, GetCent(full_rect).y), r); }

	// return default knob rect centred around knob centre
	CRect getKnobRect() const
		{ return getKnobRect(CRect(0, full_rect.top, knob_width, full_rect.bottom)); }

	// return rectangle bounds for knob centre
	CRect getSlideRect() const
		{ return CRect(slide_min, full_rect.top, slide_min+(int)slide_width, full_rect.bottom); }

	bool isInKnob(CPoint p) { return getKnobRect().pointInside(p); }

	void forceState(const MouseUpdate& mouse, State state_, float drag_scale_=1.0f)
	// call to force slider into a particular state
	{
		click_pos = mouse.pos.x;
		click_val = value;
		drag_scale = drag_scale_;
		state = state_;
	}

	bool/*true=click within rectangle*/click(const MouseUpdate& mouse, float rbutton_scale=0.1f)
	// call when there's a mouse click (default behaviour, right button moves knob at 1/10th speed)
	{
		if(!isInside(mouse)) return false;
		forceState(
			mouse,
			isInKnob(mouse) ? DRAGGING : STEPPING,
			mouse.click_mask&kLButton ? 1.0f:rbutton_scale
		);
		return true;
	}

	bool/*value changed*/update(const MouseUpdate& mouse)
	//
	// return true if value has changed
	// dragging, focus and chk for mouse up
	//
	{
		// look for mouse-up
		if(!mouse.isAnyButtonDown(/*raw*/false)) state = INACTIVE;

		// these states mean the mouse button is pressed
		switch(state) {
			case DRAGGING: {
				float delta = (float)(mouse.pos.x-click_pos)/slide_width;
				return setValue(click_val+delta*drag_scale);
			}
			case STEPPING: {
				if(mouse.pos.x > knob_cent) return setValue(value+step_speed*drag_scale);
				if(mouse.pos.x < knob_cent) return setValue(value-step_speed*drag_scale);
				return false;
			}
		}

		// if any button is down but we didn't get the click or the mouse isn't in our
		// rectangle then we are inactive
		if(mouse.isAnyButtonDown(/*raw*/true) || !isInside(mouse)) state = INACTIVE;
		else {
			// mouse is inside our rectangle, we have focus, chk whether it is over the
			// knob or not
			state = isInKnob(mouse) ? OVER_KNOB : FOCUS;
		}
		return false;
	}


	// return true if knob needs redraw because value changed or state has changed
	bool /*true=needs redraw*/ isDirty() const
		{ return val_chg || state != state_drawn; }

	void setDirty(bool dirty)
	// set or clear dirty state
	{
		val_chg = dirty;
		state_drawn = dirty?FORCE_DIRTY:state;
	}

	bool/*value changed*/setValue(float v)
	// override OwnerControl
	// set the value and update knob_rect
	{
		bool r = OwnerControl::setValue(v);
		knob_cent = slide_min+(int)(0.5f+lin_interp(value, 0.0f, slide_width));
		return r;
	}

	// increment per call to update() when "stepping"
	float step_speed;

	// treat all of these as read only from outside of struct

	// width of knob
	int knob_width;

	// centre of knob
	int knob_cent;

	//
	float drag_scale;

	// knob centre position corresponding to value 0.0f
	int slide_min;

	// width of sliding rectangle
	float slide_width;		

	// state of the slider and state last drawn
	State state, state_drawn;

	// mouse position when knob was clicked
	int click_pos;

	// value when knob was clicked
	float click_val;

};

//-------------------------------------------------------------------------------------------------
struct OwnerInvisiHSlider : public OwnerControl
//
// owner invisible horizontal slider
// dragging in x-direction changes value while y-direction changes scale factor
//
{
public:
	typedef enum {
		FORCE_DIRTY = -1,
		INACTIVE = 0,
		FOCUS,
		DRAGGING
	} State;

	OwnerInvisiHSlider() { }

	void init(
		CRect full_rect_,	// rectangle
		int xscale,		// pixels over which mouse must move for 0 to 1 (approx)
		int yscale		// number of pixels in y direction for double speed
	)
	{
		mouse_pointer = LoadCursor(/*instance*/NULL, MAKEINTRESOURCE(IDC_SIZEWE));
		old_pointer = NULL;
		drag_capture = false;

		full_rect = full_rect_;
		xmult = 1.0f/(float)xscale;
		ymult = 1.0f/(float)yscale;
		state = INACTIVE;
		setDirty(true);
	}

	void forceState(const MouseUpdate& mouse, State state_, float scale)
	// call to force slider into a particular state
	{
		state = state_;
		click_pos = mouse;
		prev_pos = mouse;
		click_val = value;
		click_mult = scale*xmult;
	}

	bool/*true=click within rectangle*/click(const MouseUpdate& mouse, float rbutton_scale=0.2f)
	//
	// call when there's a mouse click, default behaviour is to move at 1/5th normal speed
	// on rt button press
	{
		if(!isInside(mouse)) return false;
		if(state != DRAGGING) {
			drag_capture = true;
			old_pointer = SetCursor(mouse_pointer);
			HWND active_wdw = GetActiveWindow(); // hopefully that's us
			if(active_wdw) SetCapture(active_wdw);
		}
		forceState(mouse, DRAGGING, mouse.click_mask&kLButton ? 1.0f:rbutton_scale);
		return true;
	}

	bool/*value changed*/update(const MouseUpdate& mouse)
	//
	// return true if value has changed
	// dragging, focus and chk for mouse up
	//
	{

		// look for mouse-up
		if(!mouse.isAnyButtonDown(/*raw*/false)) state = INACTIVE;

		if(state == DRAGGING) {
			// calculate the multiply 
			float mult = click_mult;

			// move faster if mouse is further in y direction from click pos
			float y_offs = (float)(mouse-click_pos).y;
			mult *= fabsf(y_offs)*ymult+1.0f;

			// get x distance from previous point
			float x_dist = (float)(mouse-prev_pos).x;

			setValue(value + x_dist*mult);
			prev_pos = mouse;
			return val_chg; // updated by setValue
		}

		// don't take focus if the mouse is down & we didn't get the click or the pointer
		// is not in our rectangle
		if(mouse.isAnyButtonDown(/*raw*/true) || !isInside(mouse)) state = INACTIVE;
		else state = FOCUS;

		// chk whether we need to release the mouse after dragging
		if(state != DRAGGING && drag_capture) {
			drag_capture = false;
			ReleaseCapture();
			SetCursor(old_pointer);
			old_pointer = NULL;
		}
		return false;
	}

	// return true if knob needs redraw because value changed or state has changed
	bool /*true=needs redraw*/ isDirty() const
		{ return val_chg || state != state_drawn; }

	void setDirty(bool dirty)
	// set or clear dirty state
	{
		val_chg = dirty;
		state_drawn = dirty?FORCE_DIRTY:state;
	}

	float xmult;

	float ymult;

	//
	float click_mult;

	float click_val;

	// click position
	CPoint click_pos;

	// previous position during drag
	CPoint prev_pos;

	// state of the slider and state last drawn
	State state, state_drawn;

	// windows stuff, chg mouse pointer when being dragged
	HCURSOR mouse_pointer, old_pointer;

	//
	bool drag_capture;
};

//-------------------------------------------------------------------------------------------------
class OwnerKnob : public OwnerControl
//
// owner knob for classes that want to draw all custom graphics, turn clockwise to increase value
//
{
public:
	typedef enum {
		FORCE_DIRTY = -1,
		INACTIVE = 0,
		FOCUS,
		SPINNING
	} State;

	OwnerKnob() {}

	void init(
		CRect full_rect_,			// full rectangle size
		CPoint knob_cent_offs,		// position of knob centre relative to full_rect
		float zero_pos_frac_=0.0f	// zero-position: 0.0f=left, 0.25f=top, 0.5f=right, 0.75f=bottom
	)
	{
		zero_pos_frac = zero_pos_frac_;
		full_rect = full_rect_;
		knob_cent = knob_cent_offs+TopLeft(full_rect);
		state = INACTIVE;
		setDirty(true);
	}

	// get the knob point on the edge of a circle with specified radius, using "angle_offs"
	// if "angle_offs" is 0, value of 0 corresponds to 
	CPoint getKnobPoint(float radius) const
	{
		float angle = (value+zero_pos_frac-0.5f)*2.0f*3.1415927f;
		return knob_cent + CPoint(
			(int)(0.5f+radius*cosf(angle)),
			(int)(0.5f+radius*sinf(angle))
		);
	}

	float/*0 .. 1.0f*/getAngleFrac(CPoint p, float add) const
	// turn point into angle
	{
		p = p-knob_cent;
		if(p == CPoint(0,0)) return 0;
		float v =  atan2f((float)p.y, (float)p.x)*(0.5f/3.1415927f)+add-zero_pos_frac+0.5f;
		float dummy;
		v = modf(v, /*integer part*/&dummy);
		if(v < 0.0f) v += 1.0f;
		return v;
	}

	void forceState(const MouseUpdate& mouse, State state_, bool absolute_mode)
	// call to force slider into a particular state
	{
		state = state_;

		if(absolute_mode) {
			click_frac = 0.0f;
			setValue(getAngleFrac(mouse, /*add*/0.0f));
		}
		// otherwise relative mode
		else click_frac = value-getAngleFrac(mouse, /*add*/0.0f);
	}

	bool/*true=click within rectangle*/click(const MouseUpdate& mouse, bool absolute_mode)
	//
	// call when there's a mouse click
	//
	{
		if(!isInside(mouse)) return false;
		forceState(mouse, SPINNING, absolute_mode);
		return true;
	}

	bool/*value changed*/update(const MouseUpdate& mouse)
	//
	// return true if value has changed
	// dragging, focus and chk for mouse up
	//
	{
		// look for mouse-up
		if(!mouse.isAnyButtonDown(/*raw*/false)) state = INACTIVE;

		if(state == SPINNING) {
			float v = getAngleFrac(mouse, /*add*/click_frac);

			// don't let knob flip from one limit to another
			if(v < 0.25f && value > 0.75f) v = 1.0f;
			if(v > 0.75f && value < 0.25f) v = 0.0f;
			return setValue(v);
		}

		// don't take focus if the mouse is down & we didn't get the click or the pointer
		// is not in our rectangle
		if(mouse.isAnyButtonDown(/*raw*/true) || !isInside(mouse)) state = INACTIVE;
		else state = FOCUS;
		return false;
	}


	// return true if knob needs redraw because value changed or state has changed
	bool /*true=needs redraw*/ isDirty() const
		{ return val_chg || state != state_drawn; }

	void setDirty(bool dirty)
	// set or clear dirty state
	{
		val_chg = dirty;
		state_drawn = dirty?FORCE_DIRTY:state;
	}

	// normalized angle of click (0 .. 1)
	float click_frac;

	// angle frac offset for zero position of knob (0 .. 1)
	float zero_pos_frac;

	// centre of knob, same origin as full_rect 
	CPoint knob_cent;

	// state of the slider and state last drawn
	State state, state_drawn;
};


#endif

